queryHandler = {
            'mardi': {
                      'step':  '''PREFIX wdt:<https://portal.mardi4nfdi.de/prop/direct/>
                                  PREFIX wd:<https://portal.mardi4nfdi.de/entity/>

                                  SELECT (GROUP_CONCAT(DISTINCT(?msc); SEPARATOR=" / ") AS ?mscID)
                                         (GROUP_CONCAT(DISTINCT CONCAT(?input, " | ", ?inputl, " | ", ?inputd); separator=" / ") AS ?inputDataSet)
                                         (GROUP_CONCAT(DISTINCT CONCAT(?output, " | ", ?outputl, " | ", ?outputd); separator=" / ") AS ?outputDataSet)
                                         (GROUP_CONCAT(DISTINCT CONCAT(?method, " | ", ?methodl, " | ", ?methodd, " | ", STR(?methodurl)); separator=" / ") AS ?uses)
                                         (GROUP_CONCAT(DISTINCT CONCAT(?platformsoftware, " | ", ?platformsoftwarel, " | ", ?platformsoftwared); separator=" / ") AS ?platformSoftware)
                                         (GROUP_CONCAT(DISTINCT CONCAT(?platforminstrument, " | ", ?platforminstrumentl, " | ", ?platforminstrumentd); separator=" / ") AS ?platformInstrument)
                                         (GROUP_CONCAT(DISTINCT CONCAT(?field, " | ", ?fieldl, " | ", ?fieldd); separator=" / ") AS ?fieldOfWork)
                                  
                                  WHERE {{
                                          VALUES ?step {{ wd:{0} }}
                                          OPTIONAL {{
                                                     ?step wdt:{input data set} ?inputraw
                                                     BIND(CONCAT("mardi:", STRAFTER(STR(?inputraw), STR(wd:))) AS ?input)
                                                     
                                                     OPTIONAL {{
                                                                ?inputraw rdfs:label ?inputlraw
                                                                FILTER (lang(?inputlraw) = 'en')
                                                              }}
                                                     
                                                     BIND(COALESCE(?inputlraw, "No Label Provided!") AS ?inputl)
                                                     
                                                     OPTIONAL {{
                                                                ?inputraw schema:description ?inputdraw
                                                                FILTER (lang(?inputdraw) = 'en')
                                                              }}
                                                     
                                                     BIND(COALESCE(?inputdraw, "No Description Provided!") AS ?inputd)
                                                   }}
                                          OPTIONAL {{
                                                     ?step wdt:{output data set} ?outputraw
                                                     BIND(CONCAT("mardi:", STRAFTER(STR(?outputraw), STR(wd:))) AS ?output)
                                                     
                                                     OPTIONAL {{
                                                                ?outputraw rdfs:label ?outputlraw
                                                                FILTER (lang(?outputlraw) = 'en')
                                                              }}
                                                     
                                                     BIND(COALESCE(?outputlraw, "No Label Provided!") AS ?outputl)
                                                     
                                                     OPTIONAL {{
                                                                ?outputraw schema:description ?outputdraw
                                                                FILTER (lang(?outputdraw) = 'en')
                                                              }}
                                                     
                                                     BIND(COALESCE(?outputdraw, "No Description Provided!") AS ?outputd)
                                                   }}
                                          OPTIONAL {{
                                                     ?step wdt:{uses} ?methodraw
                                                     BIND(CONCAT("mardi:", STRAFTER(STR(?methodraw), STR(wd:))) AS ?method)
                                                     
                                                     OPTIONAL {{
                                                                ?methodraw rdfs:label ?methodlraw
                                                                FILTER (lang(?methodlraw) = 'en')
                                                              }}
                                                     
                                                     BIND(COALESCE(?methodlraw, "No Label Provided!") AS ?methodl)
                                                     
                                                     OPTIONAL {{
                                                                ?methodraw schema:description ?methoddraw
                                                                FILTER (lang(?methoddraw) = 'en')
                                                              }}
                                                     
                                                     BIND(COALESCE(?methoddraw, "No Description Provided!") AS ?methodd)
                                                     OPTIONAL {{
                                                                ?methodraw wdt:{URL} ?methodurl
                                                              }}
                                                     BIND(COALESCE(?methodurl, "") AS ?methodurl)
                                                   }}
                                          OPTIONAL {{
                                                     ?step p:{platform} ?statement0.
                                                     ?statement0 ps:{platform} ?platformsoftwareraw.
                                                    
                                                     ?statement pq:{object has role} ?platformsoftwaretyperaw.
                                                     BIND(CONCAT("", STRAFTER(STR(?platformsoftwaretyperaw), STR(wd:))) AS ?platformsoftwaretype)
                                                     FILTER (?platformsoftwaretype = "{software}")
                                                     
                                                     BIND(CONCAT("mardi:", STRAFTER(STR(?platformsoftwareraw), STR(wd:))) AS ?platformsoftware)
                                                     
                                                     OPTIONAL {{
                                                                ?platformsoftwareraw rdfs:label ?platformsoftwarelraw
                                                                FILTER (lang(?platformsoftwarelraw) = 'en')
                                                              }}
                                                     
                                                     BIND(COALESCE(?platformsoftwarelraw, "No Label Provided!") AS ?platformsoftwarel)
                                                     
                                                     OPTIONAL {{
                                                                ?platformsoftwareraw schema:description ?platformsoftwaredraw
                                                                FILTER (lang(?platformsoftwaredraw) = 'en')
                                                              }}
                                                     
                                                     BIND(COALESCE(?platformsoftwaredraw, "No Description Provided!") AS ?platformsoftwared)
                                                   }}
                                          OPTIONAL {{
                                                     ?step p:{platform} ?statement1.
                                                     ?statement1 ps:{platform} ?platforminstrumentraw.
                                                     
                                                     ?statement pq:{object has role} ?platforminstrumenttyperaw.
                                                     BIND(CONCAT("", STRAFTER(STR(?platforminstrumenttyperaw), STR(wd:))) AS ?platforminstrumenttype)
                                                     FILTER (?platforminstrumenttype = "{research tool}")
                                                     
                                                     BIND(CONCAT("mardi:", STRAFTER(STR(?platforminstrumentraw), STR(wd:))) AS ?platforminstrument)
                                                     
                                                     OPTIONAL {{
                                                                ?platforminstrumentraw rdfs:label ?platforminstrumentlraw
                                                                FILTER (lang(?platforminstrumentlraw) = 'en')
                                                              }}
                                                     
                                                     BIND(COALESCE(?platforminstrumentlraw, "No Label Provided!") AS ?platforminstrumentl)
                                                     
                                                     OPTIONAL {{
                                                                ?platforminstrumentraw schema:description ?platforminstrumentdraw
                                                                FILTER (lang(?platforminstrumentdraw) = 'en')
                                                              }}
                                                     
                                                     BIND(COALESCE(?platforminstrumentdraw, "No Description Provided!") AS ?platforminstrumentd)
                                                   }}
                                          OPTIONAL {{
                                                     ?step wdt:{field of work} ?fieldraw.
                                                     
                                                     BIND(CONCAT("mardi:", STRAFTER(STR(?fieldraw), STR(wd:))) AS ?field)
                                                     OPTIONAL {{
                                                                ?fieldraw rdfs:label ?fieldlraw
                                                                FILTER (lang(?fieldlraw) = 'en')
                                                              }}
                                                     
                                                     BIND(COALESCE(?fieldlraw, "No Label Provided!") AS ?fieldl)
                                                     
                                                     OPTIONAL {{
                                                                ?fieldraw schema:description ?fielddraw
                                                                FILTER (lang(?fielddraw) = 'en')
                                                              }}
                                                     
                                                     BIND(COALESCE(?fielddraw, "No Description Provided!") AS ?fieldd)
                                                   }}
                                          OPTIONAL {{
                                                     ?step wdt:{MSC ID} ?msc.
                                                   }}
                                  }}''',

                         'method':  '''PREFIX wdt:<https://portal.mardi4nfdi.de/prop/direct/>
                                       PREFIX wd:<https://portal.mardi4nfdi.de/entity/>
     
                                       SELECT (GROUP_CONCAT(DISTINCT CONCAT(?impSoft, " | ", ?impSoftl, " | ", ?impSoftd); separator=" / ") AS ?implementedBySoftware)
                                              (GROUP_CONCAT(DISTINCT CONCAT(?impInst, " | ", ?impInstl, " | ", ?impInstd); separator=" / ") AS ?implementedByInstrument)
                                              
                                       WHERE {{
                                               VALUES ?method {{ wd:{0} }}
                                               OPTIONAL {{
                                                          ?method wdt:P ?impSoftraw.
                                                          ?impSoftraw wdt:{instance of} wd:{software}.

                                                          BIND(CONCAT("mardi:", STRAFTER(STR(?impSoftraw), STR(wd:))) AS ?impSoft)
                                                          
                                                          OPTIONAL {{
                                                                     ?impSoftraw rdfs:label ?impSoftlraw
                                                                     FILTER (lang(?impSoftlraw) = 'en')
                                                                   }}
                                                          
                                                          BIND(COALESCE(?impSoftlraw, "No Label Provided!") AS ?impSoftl)
                                                          
                                                          OPTIONAL {{
                                                                     ?impSoftraw schema:description ?impSoftdraw
                                                                     FILTER (lang(?impSoftdraw) = 'en')
                                                                   }}
                                                          
                                                          BIND(COALESCE(?impSoftdraw, "No Description Provided!") AS ?impSoftd)
                                                        }}
                                                        
                                              OPTIONAL {{
                                                          ?method wdt:P ?impInstraw.
                                                          ?impInstraw wdt:{instance of} wd:{research tool}.
                                                          
                                                          BIND(CONCAT("mardi:", STRAFTER(STR(?impInstraw), STR(wd:))) AS ?impInst)
                                                          
                                                          OPTIONAL {{
                                                                     ?impInstraw rdfs:label ?impInstlraw
                                                                     FILTER (lang(?impInstlraw) = 'en')
                                                                   }}
                                                          
                                                          BIND(COALESCE(?impInstlraw, "No Label Provided!") AS ?impInstl)
                                                          
                                                          OPTIONAL {{
                                                                     ?impInstraw schema:description ?impInstdraw
                                                                     FILTER (lang(?impInstdraw) = 'en')
                                                                   }}
                                                          
                                                          BIND(COALESCE(?impInstdraw, "No Description Provided!") AS ?impInstd)
                                                        }}
                                            }}''',

                         'software': '''PREFIX wdt:<https://portal.mardi4nfdi.de/prop/direct/>
                                        PREFIX wd:<https://portal.mardi4nfdi.de/entity/>
        
                                        SELECT ?userManualURL ?sourceCodeRepository ?reference
                                               (GROUP_CONCAT(DISTINCT CONCAT(?pl, " | ", ?pll, " | ", ?pld); separator=" / ") AS ?programmedIn)
                                               (GROUP_CONCAT(DISTINCT CONCAT(?dp, " | ", ?dpl, " | ", ?dpd); separator=" / ") AS ?dependsOnSoftware)
               
                                        WHERE {{
                                          
                                          VALUES ?software {{ wd:{0} }}
                                          
                                          # Get Programming Language
                                          OPTIONAL {{
                                                     ?software wdt:{programmed in} ?plraw.
                                                     BIND(CONCAT("mardi:", STRAFTER(STR(?plraw), STR(wd:))) AS ?pl)
                                                     
                                                     OPTIONAL {{
                                                                ?plraw rdfs:label ?pllraw
                                                                FILTER (lang(?pllraw) = 'en')
                                                              }}
                                                     
                                                     BIND(COALESCE(?pllraw, "No Label Provided!") AS ?pll)
                                                     
                                                     OPTIONAL {{
                                                                ?plraw schema:description ?pldraw
                                                                FILTER (lang(?pldraw) = 'en')
                                                              }}
                                                     
                                                     BIND(COALESCE(?pldraw, "No Description Provided!") AS ?pld)
                                                   }}  
                                          
                                          # Get Dependencies
                                          OPTIONAL {{
                                                     ?software wdt:{depends on software} ?dpraw.
                                                     BIND(CONCAT("mardi:", STRAFTER(STR(?dpraw), STR(wd:))) AS ?dp)
                                                     
                                                     OPTIONAL {{
                                                                ?dpraw rdfs:label ?dplraw
                                                                FILTER (lang(?dplraw) = 'en')
                                                              }}
                                                     
                                                     BIND(COALESCE(?dplraw, "No Label Provided!") AS ?dpl)
                                                     
                                                     OPTIONAL {{
                                                                ?dpraw schema:description ?dpdraw
                                                                FILTER (lang(?dpdraw) = 'en')
                                                              }}
                                                     
                                                     BIND(COALESCE(?dpdraw, "No Description Provided!") AS ?dpd)
                                                   }}
        
                                          # Source Code Repository
                                          OPTIONAL {{
                                                     ?software wdt:{source code repository URL} ?sourceCodeRepository.
                                                   }}
        
                                          # User Manual URL
                                          OPTIONAL {{
                                                     ?software wdt:{user manual URL} ?userManualURL.
                                                   }}
                                          # DOI
                                          OPTIONAL {{
                                                     ?software wdt:{DOI} ?doi.
                                                   }}
                                          
                                          # SWMATH
                                          OPTIONAL {{
                                                     ?software wdt:{swMath work ID} ?swmath.
                                                   }}

                                          # URL
                                          OPTIONAL {{
                                                     ?software wdt:{URL} ?url.
                                                   }}

                                         BIND(
                                              CONCAT(
                                                  IF(BOUND(?doi), CONCAT("doi:", STR(?doi), " | "), ""),
                                                  IF(BOUND(?swmath), CONCAT("swmath:", STR(?swmath), " | "), ""),
                                                  IF(BOUND(?url), STR(?url), "")
                                              ) AS ?referenceraw
                                          )
                                          
                                          BIND(IF(STRENDS(?referenceraw, " | "), STRBEFORE(?referenceraw, " | "), ?referenceraw) AS ?reference)

                                        }}  
                                        
                                        GROUP BY ?sourceCodeRepository ?userManualURL ?reference''',

                         'hardware': '''PREFIX wdt:<https://portal.mardi4nfdi.de/prop/direct/>
                                        PREFIX wd:<https://portal.mardi4nfdi.de/entity/>
        
                                        SELECT ?nodes ?cores
                                               (GROUP_CONCAT(DISTINCT CONCAT(?cpu, " | ", ?cpul, " | ", ?cpud); separator=" / ") AS ?CPU)
               
                                        WHERE {{

                                          VALUES ?hardware {{ wd:{0} }}
        
                                          # Get CPU
                                          OPTIONAL {{
                                                     ?hardware p:{CPU} ?statement.
                                                     ?statement ps:{CPU} ?cpuraw.
                                                     BIND(CONCAT("mardi:", STRAFTER(STR(?cpuraw), STR(wd:))) AS ?cpu)
                                                     
                                                     OPTIONAL {{
                                                                ?cpuraw rdfs:label ?cpulraw
                                                                FILTER (lang(?cpulraw) = 'en')
                                                              }}
                                                     
                                                     BIND(COALESCE(?cpulraw, "No Label Provided!") AS ?cpul)
                                                     
                                                     OPTIONAL {{
                                                                ?cpuraw schema:description ?cpudraw
                                                                FILTER (lang(?cpudraw) = 'en')
                                                              }}
                                                     
                                                     BIND(COALESCE(?cpudraw, "No Description Provided!") AS ?cpud)
        
                                                     OPTIONAL {{
                                                                ?cpuraw wdt:{number of processor cores} ?cores.
                                                              }}
                                                                
                                                     OPTIONAL {{
                                                                ?statement pq:{quantity_property} ?nodes.
                                                              }}
                                                   }}
                                              }}
                                        GROUP BY ?nodes ?cores ''',

                         'data-set':  '''PREFIX wdt:<https://portal.mardi4nfdi.de/prop/direct/>
                                         PREFIX wd:<https://portal.mardi4nfdi.de/entity/>
         
                                         SELECT ?sizeValue ?sizeUnit ?sizeRecord ?fileFormat ?binaryOrText ?proprietary
                                                ?DOI ?URL ?publish ?archive ?endTime
                                                (GROUP_CONCAT(DISTINCT CONCAT(?datatype, " | ", ?datatypel, " | ", ?datatyped); separator=" / ") AS ?dataType)
                                                (GROUP_CONCAT(DISTINCT CONCAT(?representationformat, " | ", ?representationformatl, " | ", ?representationformatd); separator=" / ") AS ?representationFormat)
                
                                         WHERE {{

                                           VALUES ?dataset {{ wd:{0} }}
         
                                           # Get Size (bytes)
                                           OPTIONAL {{
                                                      ?dataset p:{data size} ?statementNode.          
                                                      ?statementNode psv:{data size} ?valueNode.      
                                                      ?valueNode wikibase:quantityAmount ?sizeValue; 
                                                                 wikibase:quantityUnit ?unit.
                                                      ?unit rdfs:label ?sizeUnit.
                                                      FILTER (lang(?sizeUnit) = 'en')
                                                    }}

                                           # Get Size (number of items)
                                           OPTIONAL {{
                                                      ?dataset wdt:{number of records} ?sizeRecord
                                                    }}
         
                                           # Data Type
                                           OPTIONAL {{
                                                      ?dataset p:{uses} ?statement_dt.
                                                      ?statement_dt ps:{uses} ?datatyperaw. 
                                                      ?statement_dt pq:{object has role} ?qualifier.
                                                      FILTER (?qualifier = wd:{data type})
         
                                                      BIND(CONCAT("mardi:", STRAFTER(STR(?datatyperaw), STR(wd:))) AS ?datatype)
                                                      
                                                      OPTIONAL {{
                                                                 ?datatyperaw rdfs:label ?datatypelraw
                                                                 FILTER (lang(?datatypelraw) = 'en')
                                                               }}
                                                      
                                                      BIND(COALESCE(?datatypelraw, "No Label Provided!") AS ?datatypel)
                                                      
                                                      OPTIONAL {{
                                                                 ?datatyperaw schema:description ?datatypedraw
                                                                 FILTER (lang(?datatypedraw) = 'en')
                                                               }}
                                                      
                                                      BIND(COALESCE(?datatypedraw, "No Description Provided!") AS ?datatyped)
                                                    }}  
           
                                           #Representation Format
                                           OPTIONAL {{
                                                      ?dataset p:{uses} ?statement_dt2.
                                                      ?statement_dt2 ps:{uses} ?representationformatraw. 
                                                      ?statement_dt2 pq:{object has role} ?qualifier2.
                                                      FILTER (?qualifier2 = wd:{representation format})
         
                                                      BIND(CONCAT("mardi:", STRAFTER(STR(?representationformatraw), STR(wd:))) AS ?representationformat)
                                                      
                                                      OPTIONAL {{
                                                                 ?representationformatraw rdfs:label ?representationformatlraw
                                                                 FILTER (lang(?representationformatlraw) = 'en')
                                                               }}
                                                      
                                                      BIND(COALESCE(?representationformatlraw, "No Label Provided!") AS ?representationformatl)
                                                      
                                                      OPTIONAL {{
                                                                 ?representationformatraw schema:description ?representationformatdraw
                                                                 FILTER (lang(?representationformatdraw) = 'en')
                                                               }}
                                                      
                                                      BIND(COALESCE(?representationformatdraw, "No Description Provided!") AS ?representationformatd)
                                                    }}
         
                                           # Get file extension
                                           OPTIONAL {{
                                                      ?dataset wdt:{file extension} ?fileFormat
                                                    }}
                                               
                                           # Binary or Text
                                           BIND(IF(EXISTS {{ ?dataset wdt:{instance of} wd:{binary data} }}, "binary", IF(EXISTS {{ ?dataset wdt:{instance of} wd:{text data} }}, "text", "" )) AS ?binaryOrText)
         
                                           # Proprietary
                                           BIND(IF(EXISTS {{ ?dataset wdt:{instance of} wd:{proprietary information} }}, "Yes", IF(EXISTS {{ ?dataset wdt:{instance of} wd:{open data} }}, "No", "" )) AS ?proprietary)
                                           
                                           # Bind DOI and URL if they exist
                                           OPTIONAL {{ ?dataset wdt:{DOI} ?DOI. }}
                                           OPTIONAL {{ ?dataset wdt:{URL} ?URL. }}
         
                                           # Data Publishing
                                           BIND(IF(EXISTS {{ ?dataset wdt:{mandates} wd:{data publishing} }}, "Yes", "No") AS ?publish)
         
                                           # Data Archiving
                                           BIND(IF(EXISTS {{ ?dataset wdt:{mandates} wd:{research data archiving} }}, "YesText", "NoText") AS ?archive)
         
                                           OPTIONAL {{
                                                      ?dataset p:{mandates} ?statementNode2.
                                                      ?statementNode2 ps:{mandates} wd:{research data archiving}.
                                                      OPTIONAL {{
                                                                  ?statementNode2 pq:{end time} ?endTime.
                                                               }}
                                                    }}
                                           }}    
                                           GROUP BY ?sizeValue ?sizeUnit ?sizeRecord ?fileFormat ?binaryOrText ?proprietary ?DOI ?URL ?publish ?archive ?endTime'''
                     },
            'wikidata': { 
                         'step':  '''SELECT (GROUP_CONCAT(DISTINCT(?msc); SEPARATOR=" / ") AS ?mscID)
                                            (GROUP_CONCAT(DISTINCT CONCAT(?input, " | ", ?inputl, " | ", ?inputd); separator=" / ") AS ?inputDataSet)
                                            (GROUP_CONCAT(DISTINCT CONCAT(?output, " | ", ?outputl, " | ", ?outputd); separator=" / ") AS ?outputDataSet)
                                            (GROUP_CONCAT(DISTINCT CONCAT(?method, " | ", ?methodl, " | ", ?methodd); separator=" / ") AS ?uses)
                                            (GROUP_CONCAT(DISTINCT CONCAT(?platformsoftware, " | ", ?platformsoftwarel, " | ", ?platformsoftwared); separator=" / ") AS ?platformSoftware)
                                            (GROUP_CONCAT(DISTINCT CONCAT(?platforminstrument, " | ", ?platforminstrumentl, " | ", ?platforminstrumentd); separator=" / ") AS ?platformInstrument)
                                            (GROUP_CONCAT(DISTINCT CONCAT(?field, " | ", ?fieldl, " | ", ?fieldd); separator=" / ") AS ?fieldOfWork)

                                      WHERE {{

                                              VALUES ?step {{ wd:{0} }}

                                              OPTIONAL {{
                                                         ?step wdt:P ?inputraw
                                                         BIND(replace( xsd:string(?inputraw),'http://www.wikidata.org/entity/','wikidata:') as ?input) 

                                                         SERVICE wikibase:label {{
                                                           bd:serviceParam wikibase:language "en,[AUTO_LANGUAGE],mul" .
                                                           ?inputraw rdfs:label ?inputlraw .
                                                           ?inputraw schema:description ?inputdraw .
                                                         }}
                                                         
                                                         BIND(COALESCE(?inputlraw, "No Label Provided!") AS ?inputl)
                                                         BIND(COALESCE(?inputdraw, "No Description Provided!") AS ?inputd)

                                                       }}

                                              OPTIONAL {{
                                                         ?step wdt:P ?outputraw
                                                         BIND(replace( xsd:string(?outputraw),'http://www.wikidata.org/entity/','wikidata:') as ?output) 

                                                         SERVICE wikibase:label {{
                                                           bd:serviceParam wikibase:language "en,[AUTO_LANGUAGE],mul" .
                                                           ?outputraw rdfs:label ?outputlraw .
                                                           ?outputraw schema:description ?outputdraw .
                                                         }}
                                                         
                                                         BIND(COALESCE(?outputlraw, "No Label Provided!") AS ?outputl)
                                                         BIND(COALESCE(?outputdraw, "No Description Provided!") AS ?outputd)
                                                       }}

                                              OPTIONAL {{
                                                         ?step wdt:P2283 ?methodraw
                                                         BIND(replace( xsd:string(?methodraw),'http://www.wikidata.org/entity/','wikidata:') as ?method) 

                                                         SERVICE wikibase:label {{
                                                           bd:serviceParam wikibase:language "en,[AUTO_LANGUAGE],mul" .
                                                           ?methodraw rdfs:label ?methodlraw .
                                                           ?methodraw schema:description ?methoddraw .
                                                         }}
                                                         
                                                         BIND(COALESCE(?methodlraw, "No Label Provided!") AS ?methodl)
                                                         BIND(COALESCE(?methoddraw, "No Description Provided!") AS ?methodd)
                                                       }}

                                              OPTIONAL {{
                                                         ?step p:P400 ?statement0.
                                                         ?statement0 ps:P400 ?platformsoftwareraw.


                                                         ?statement pq:P3831 ?platformsoftwaretyperaw.
                                                         BIND(replace( xsd:string(?platformsoftwaretyperaw),'http://www.wikidata.org/entity/','') as ?platformsoftwaretype)
                                                         FILTER (?platformsoftwaretype = "Q7397")


                                                         BIND(replace( xsd:string(?platformsoftwareraw),'http://www.wikidata.org/entity/','wikidata:') as ?platformsoftware) 

                                                         SERVICE wikibase:label {{
                                                           bd:serviceParam wikibase:language "en,[AUTO_LANGUAGE],mul" .
                                                           ?platformsoftwareraw rdfs:label ?platformsoftwarelraw .
                                                           ?platformsoftwareraw schema:description ?platformsoftwaredraw .
                                                         }}
                                                         
                                                         BIND(COALESCE(?platformsoftwarelraw, "No Label Provided!") AS ?platformsoftwarel)
                                                         BIND(COALESCE(?platformsoftwaredraw, "No Description Provided!") AS ?platformsoftwared)
                                                       }}

                                              OPTIONAL {{
                                                         ?step p:P400 ?statement1.
                                                         ?statement1 ps:P400 ?platforminstrumentraw.


                                                         ?statement pq:P3831 ?platforminstrumenttyperaw.
                                                         BIND(replace( xsd:string(?platforminstrumenttyperaw),'http://www.wikidata.org/entity/','') as ?platforminstrumenttype)
                                                         FILTER (?platforminstrumenttype = "")


                                                         BIND(replace( xsd:string(?platforminstrumentraw),'http://www.wikidata.org/entity/','wikidata:') as ?platforminstrument) 

                                                         SERVICE wikibase:label {{
                                                           bd:serviceParam wikibase:language "en,[AUTO_LANGUAGE],mul" .
                                                           ?platforminstrumentraw rdfs:label ?platforminstrumentlraw .
                                                           ?platformsoftwareraw schema:description ?platforminstrumentdraw .
                                                         }}
                                                         
                                                         BIND(COALESCE(?platforminstrumentlraw, "No Label Provided!") AS ?platforminstrumentl)
                                                         BIND(COALESCE(?platforminstrumentdraw, "No Description Provided!") AS ?platforminstrumentd)
                                                       }}

                                              OPTIONAL {{
                                                         ?step wdt:P101 ?fieldraw.
                                                         BIND(replace( xsd:string(?fieldraw),'http://www.wikidata.org/entity/','wikidata:') as ?field)

                                                         SERVICE wikibase:label {{
                                                           bd:serviceParam wikibase:language "en,[AUTO_LANGUAGE],mul" .
                                                           ?fieldraw rdfs:label ?fieldlraw .
                                                           ?fieldraw schema:description ?fielddraw .
                                                         }}
                                                         
                                                         BIND(COALESCE(?fieldlraw, "No Label Provided!") AS ?fieldl)
                                                         BIND(COALESCE(?fielddraw, "No Description Provided!") AS ?fieldd)
                                                       }}

                                              OPTIONAL {{
                                                         ?step wdt:P3285 ?msc.
                                                       }}

                                  }}''',

                           'method':  '''SELECT (GROUP_CONCAT(DISTINCT CONCAT(?impSoft, " | ", ?impSoftl, " | ", ?impSoftd); separator=" / ") AS ?implementedBySoftware)
                                                (GROUP_CONCAT(DISTINCT CONCAT(?impInst, " | ", ?impInstl, " | ", ?impInstd); separator=" / ") AS ?implementedByInstrument)
                                                
                                         WHERE {{
                                                 VALUES ?method {{ wd:{0} }}
                                                 OPTIONAL {{
                                                            ?method wdt:P ?impSoftraw.
                                                            ?impSoftraw wdt:P31 wd:Q7397.
  
                                                            BIND(CONCAT("wikidata:", STRAFTER(STR(?impSoftraw), STR(wd:))) AS ?impSoft)
                                                            
                                                            SERVICE wikibase:label {{
                                                              bd:serviceParam wikibase:language "en,[AUTO_LANGUAGE],mul" .
                                                              ?impSoftraw rdfs:label ?impSoftlraw .
                                                              ?impSoftraw schema:description ?impSoftdraw .
                                                            }}
                                                            
                                                            BIND(COALESCE(?impSoftlraw, "No Label Provided!") AS ?impSoftl)
                                                            BIND(COALESCE(?impSoftdraw, "No Description Provided!") AS ?impSoftd)
                                                          }}
                                                          
                                                OPTIONAL {{
                                                            ?method wdt:P ?impInstraw.
                                                            ?impInstraw wdt:P31 wd:Q110994345.
                                                            
                                                            BIND(CONCAT("wikidata:", STRAFTER(STR(?impInstraw), STR(wd:))) AS ?impInst)
                                                            
                                                            SERVICE wikibase:label {{
                                                              bd:serviceParam wikibase:language "en,[AUTO_LANGUAGE],mul" .
                                                              ?impInstraw rdfs:label ?impInstlraw .
                                                              ?impInstraw schema:description ?impInstdraw .
                                                            }}
                                                            
                                                            BIND(COALESCE(?impInstlraw, "No Label Provided!") AS ?impInstl)
                                                            BIND(COALESCE(?impInstdraw, "No Description Provided!") AS ?impInstd)
                                                          }}
                                               }}''',

                           'software':  '''SELECT ?sourceCodeRepository ?userManualURL ?reference
                                                  (GROUP_CONCAT(DISTINCT CONCAT(?pl, " | ", ?pll, " | ", ?pld); separator=" / ") AS ?programmedIn)
                                                  (GROUP_CONCAT(DISTINCT CONCAT(?dp, " | ", ?dpl, " | ", ?dpd); separator=" / ") AS ?dependsOnSoftware)
                                       
                                           WHERE {{

                                             VALUES ?software {{ wd:{0} }}
           
                                             # Get Programming Language
                                             OPTIONAL {{
                                                        ?software wdt:P277 ?plraw.
                                                        BIND(replace( xsd:string(?plraw),'http://www.wikidata.org/entity/','wikidata:') as ?pl) 
                                                        
                                                        SERVICE wikibase:label {{
                                                              bd:serviceParam wikibase:language "en,[AUTO_LANGUAGE],mul" .
                                                              ?plraw rdfs:label ?pllraw .
                                                              ?plraw schema:description ?pldraw .
                                                            }}
                                                            
                                                            BIND(COALESCE(?pllraw, "No Label Provided!") AS ?pll)
                                                            BIND(COALESCE(?pldraw, "No Description Provided!") AS ?pld)
                                                      }}  
                                             
                                             # Get Dependencies
                                             OPTIONAL {{
                                                        ?software wdt:P1547 ?dpraw.
                                                        BIND(replace( xsd:string(?dpraw),'http://www.wikidata.org/entity/','wikidata:') as ?dp) 
                                                        
                                                        SERVICE wikibase:label {{
                                                              bd:serviceParam wikibase:language "en,[AUTO_LANGUAGE],mul" .
                                                              ?dpraw rdfs:label ?dplraw .
                                                              ?dpraw schema:description ?dpdraw .
                                                            }}
                                                            
                                                            BIND(COALESCE(?dplraw, "No Label Provided!") AS ?dpl)
                                                            BIND(COALESCE(?dpdraw, "No Description Provided!") AS ?dpd)
                                                      }}
           
                                             # Source Code Published
                                             OPTIONAL {{
                                                        ?software wdt:P1324 ?published.
                                                      }}
           
                                             # User Manual Documented
                                             OPTIONAL {{
                                                        ?software wdt:P2078 ?documented.
                                                      }}
                                             # DOI
                                             OPTIONAL {{
                                                        ?software wdt:P356 ?doi.
                                                      }}
                                             
                                             # SWMATH
                                             OPTIONAL {{
                                                        ?software wdt:P6830 ?swmath.
                                                      }}

                                             # URL
                                             OPTIONAL {{
                                                        ?software wdt:P2699 ?url.
                                                      }}

                                             BIND(
                                              CONCAT(
                                                  IF(BOUND(?doi), CONCAT("doi:", STR(?doi), " | "), ""),
                                                  IF(BOUND(?swmath), CONCAT("swmath:", STR(?swmath), " | "), ""),
                                                  IF(BOUND(?url), STR(?url), "")
                                              ) AS ?referenceraw
                                             )
                                          
                                             BIND(IF(STRENDS(?referenceraw, " | "), STRBEFORE(?referenceraw, " | "), ?referenceraw) AS ?reference)
           
                                            }}
                                           GROUP BY ?sourceCodeRepository ?userManualURL ?reference''',

                           'hardware': '''SELECT ?nodes ?cores
                                          (GROUP_CONCAT(DISTINCT CONCAT(?cpu, " | ", ?cpul, " | ", ?cpud); separator=" / ") AS ?CPU)
       
                                          WHERE {{

                                            VALUES ?hardware {{ wd:{0} }}

                                            # Get CPU
                                            OPTIONAL {{
                                                       ?hardware p:P880 ?statement.
                                                       ?statement ps:P880 ?cpuraw.
                                                       BIND(replace( xsd:string(?cpuraw),'http://www.wikidata.org/entity/','wikidata:') as ?cpu) 

                                                       SERVICE wikibase:label {{
                                                         bd:serviceParam wikibase:language "en,[AUTO_LANGUAGE],mul" .
                                                         ?cpuraw rdfs:label ?cpulraw .
                                                         ?cpuraw schema:description ?cpudraw .
                                                       }}
                                                       
                                                       BIND(COALESCE(?cpulraw, "No Label Provided!") AS ?cpul)
                                                       BIND(COALESCE(?cpudraw, "No Description Provided!") AS ?cpud)

                                                       OPTIONAL {{
                                                                  ?cpuraw wdt:P1141 ?cores.
                                                                }}

                                                       OPTIONAL {{
                                                                  ?statement pq:P1114 ?nodes.
                                                                }}
                                                     }}
                                                }}
                                                GROUP BY ?nodes ?cores ''',

                           'data-set':'''SELECT ?sizeValue ?sizeUnit ?sizeRecord ?fileFormat ?binaryOrText ?proprietary
                                                ?DOI ?URL ?publish ?archive ?endTime
                                                (GROUP_CONCAT(DISTINCT CONCAT(?datatype, " | ", ?datatypel, " | ", ?datatyped); separator=" / ") AS ?dataType)
                                                (GROUP_CONCAT(DISTINCT CONCAT(?representationformat, " | ", ?representationformatl, " | ", ?representationformatd); separator=" / ") AS ?representationFormat)
       
                                         WHERE {{

                                           VALUES ?dataset {{ wd:{0} }}
         
                                           # Get Size (bytes)
                                           OPTIONAL {{
                                                      ?dataset p:P3575 ?statementNode.          
                                                      ?statementNode psv:P3575 ?valueNode.      
                                                      ?valueNode wikibase:quantityAmount ?sizeValue; 
                                                                 wikibase:quantityUnit ?unit.
                                                      ?unit rdfs:label ?sizeUnit.
                                                      FILTER (lang(?sizeUnit) = 'en')
                                                    }}

                                           # Get Size (number of items)
                                           OPTIONAL {{
                                                      ?dataset wdt:P4876 ?sizeRecord
                                                    }}
         
                                           # Data Type
                                           OPTIONAL {{
                                                      ?dataset p:P2283 ?statement_dt.
                                                      ?statement_dt ps:P2283 ?datatyperaw. 
                                                      ?statement_dt pq:P3831 ?qualifier.
                                                      FILTER (?qualifier = wd:Q190087)
         
                                                      BIND(replace( xsd:string(?datatyperaw),'http://www.wikidata.org/entity/','wikidata:') as ?datatype)
                                                      
                                                      SERVICE wikibase:label {{
                                                         bd:serviceParam wikibase:language "en,[AUTO_LANGUAGE],mul" .
                                                         ?datatyperaw rdfs:label ?datatypelraw .
                                                         ?datatyperaw schema:description ?datatypedraw .
                                                       }}
                                                       
                                                       BIND(COALESCE(?datatypelraw, "No Label Provided!") AS ?datatypel)
                                                       BIND(COALESCE(?datatypedraw, "No Description Provided!") AS ?datatyped)
                                                    }}  
           
                                           #Representation Format
                                           OPTIONAL {{
                                                      ?dataset p:P2283 ?statement_dt2.
                                                      ?statement_dt2 ps:P2283 ?representationformatraw. 
                                                      ?statement_dt2 pq:P3831 ?qualifier2.
                                                      FILTER (?qualifier2 = wd:)
         
                                                      BIND(replace( xsd:string(?representationformatraw),'http://www.wikidata.org/entity/','wikidata:') as ?representationformat)
                                                      
                                                      SERVICE wikibase:label {{
                                                         bd:serviceParam wikibase:language "en,[AUTO_LANGUAGE],mul" .
                                                         ?representationformatraw rdfs:label ?representationformatlraw .
                                                         ?representationformatraw schema:description ?representationformatdraw .
                                                       }}
                                                       
                                                       BIND(COALESCE(?representationformatlraw, "No Label Provided!") AS ?representationformatl)
                                                       BIND(COALESCE(?representationformatdraw, "No Description Provided!") AS ?representationformatd)
                                                    }}
         
                                           # Get file extension
                                           OPTIONAL {{
                                                      ?dataset wdt:P1195 ?fileFormat
                                                    }}
                                               
                                           # Binary or Text
                                           BIND(IF(EXISTS {{ ?dataset wdt:P31 wd:Q4913888 }}, "binary", IF(EXISTS {{ ?dataset wdt:P31 wd:Q60476328 }}, "text", "" )) AS ?binaryOrText)
         
                                           # Proprietary
                                           BIND(IF(EXISTS {{ ?dataset wdt:P31 wd:Q123684347 }}, "Yes", IF(EXISTS {{ ?dataset wdt:P31 wd:Q309901 }}, "No", "" )) AS ?proprietary)
                                           
                                           # Bind DOI and URL if they exist
                                           OPTIONAL {{ ?dataset wdt:P356 ?DOI. }}
                                           OPTIONAL {{ ?dataset wdt:P2699 ?URL. }}
         
                                           # Data Publishing
                                           BIND(IF(EXISTS {{ ?dataset wdt:P4424 wd:Q17051824 }}, "Yes", "No") AS ?publish)
         
                                           # Data Archiving
                                           BIND(IF(EXISTS {{ ?dataset wdt:P4424 wd:Q17155735 }}, "YesText", "NoText") AS ?archive)
         
                                           OPTIONAL {{ ?dataset p:P4424 ?statementNode2.
                                                       ?statementNode2 ps:P4424 wd:Q17155735.
                                                       OPTIONAL {{
                                                                   ?statementNode2 pq:P582 ?endTime.
                                                                }}
                                                    }}
                                           }}    
                                           GROUP BY ?sizeValue ?sizeUnit ?sizeRecord ?fileFormat ?binaryOrText ?proprietary ?DOI ?URL ?publish ?archive ?endTime'''
            },
            'mathalgodb': {
                           'method': '''PREFIX prop: <https://mardi4nfdi.de/mathalgodb/0.1#>
                                        PREFIX : <https://mardi4nfdi.de/mathalgodb/0.1/algorithm#>
                                        PREFIX rdfs: <http://www.w3.org/2000/01/rdf-schema#>                          
                                        
                                        SELECT DISTINCT (GROUP_CONCAT(DISTINCT CONCAT(?imple, " | ", ?implel, " | ", ?impled); separator=" / ") AS ?implementedBySoftware)
                                                        
                                        WHERE {{
                                                VALUES ?idraw {{ :{0} }}
         
                                                OPTIONAL {{
                                                           ?idraw prop:implementedBy ?impleraw.
                                                           BIND(CONCAT("mathalgodb:", STRAFTER(STR(?impleraw), "#")) AS ?imple)
                                                           OPTIONAL {{ ?impleraw rdfs:label ?implelraw}}
                                                           BIND(COALESCE(?implelraw, "No Label Provided!") AS ?implel)
                                                           OPTIONAL {{ ?impleraw rdfs:comment ?impledraw}}
                                                           BIND(COALESCE(?impledraw, "No Description Provided!") AS ?impled)
                                                         }}
                                              }}''',

                           'software': '''PREFIX : <https://mardi4nfdi.de/mathalgodb/0.1/software#>
                                          PREFIX dc: <http://purl.org/spar/datacite/>   
                                          
                                          SELECT DISTINCT (GROUP_CONCAT(DISTINCT ?referenceraw; separator=" | ") AS ?reference)
                                                          
                                          WHERE {{
                                                  VALUES ?idraw {{ :{0} }}
           
                                                  OPTIONAL {{ ?idraw dc:hasIdentifier ?referenceraw }}
           
                                                }}
                                          GROUP BY ?reference''' 
            }

}

queryProvider = {
                 'RT': '''SELECT DISTINCT (GROUP_CONCAT(DISTINCT CONCAT(?ta, " | ", ?tal, " | ", ?tad); separator=" / ") AS ?usedBy)
                                                       
                                       WHERE {{
                                         
                                          VALUES ?id {{wd:{0}}}

                                          OPTIONAL {{
                                                         ?id wdt:{used by} ?taraw.
                                                         BIND(CONCAT("mardi:", STRAFTER(STR(?taraw), STR(wd:))) AS ?ta)
                                                         
                                                         ?taraw wdt:{instance of} wd:{computational task}.

                                                         OPTIONAL {{
                                                                    ?taraw rdfs:label ?talraw.
                                                                    FILTER (lang(?talraw) = 'en')
                                                                 }}

                                                        BIND(COALESCE(?talraw, "No Label Provided!") AS ?tal)

                                                        OPTIONAL {{
                                                                   ?taraw schema:description ?tadraw
                                                                   FILTER (lang(?tadraw) = 'en')
                                                                 }}
                                                        BIND(COALESCE(?tadraw, "No Description Provided!") AS ?tad)
                                                     }}
                                            }}'''
}

queryPreview = {'basic': '''SELECT DISTINCT ?isSpaceContinuous ?isSpaceDiscrete
                                            ?isTimeContinuous ?isTimeDiscrete
                                                                 
                            WHERE {{
                              
                               VALUES ?id {{wd:{0}}}
                               
                               BIND(IF(EXISTS {{ ?id wdt:{instance of} wd:{continuous-space model} }}, "True", "False" ) AS ?isSpaceContinuous)
                               BIND(IF(EXISTS {{ ?id wdt:{instance of} wd:{discrete-space model} }}, "True", "False" ) AS ?isSpaceDiscrete)
                               
                               BIND(IF(EXISTS {{ ?id wdt:{instance of} wd:{continuous-time model} }}, "True", "False" ) AS ?isTimeContinuous)
                               BIND(IF(EXISTS {{ ?id wdt:{instance of} wd:{discrete-time model} }}, "True", "False" ) AS ?isTimeDiscrete)
                
                            }}
                            GROUP BY ?isSpaceContinuous ?isTimeContinuous ?isSpaceDiscrete ?isTimeDiscrete''',
                
                'variables': '''SELECT ?ID ?Name ?Unit ?Symbol ?label ?Type

                                WHERE {{ 

                                VALUES ?task {{{0}}}

                                OPTIONAL {{ ?task rdfs:label ?labelraw .
                                            FILTER (lang(?labelraw) = 'en') }}
                                BIND(COALESCE(?labelraw, "No Label Provided!") AS ?label)

                                # Define type based on connection type (Input or Output)
                                {{
                                    ?task p:{contains} ?statement.
                                    ?statement ps:{contains} ?idraw.
                                    ?statement pq:{object has role} ?role.
                                    FILTER (?role = wd:{input})
                                    BIND(CONCAT("mardi:", STRAFTER(STR(?idraw), STR(wd:))) AS ?ID)
                                    OPTIONAL {{ ?idraw rdfs:label ?Nameraw .
                                                FILTER (lang(?Nameraw) = 'en') }}
                                    BIND(COALESCE(?Nameraw, "No Label Provided!") AS ?Name)
                                    BIND("independent" AS ?Type)
                                }}
                                UNION
                                {{
                                    ?task p:{contains} ?statement.
                                    ?statement ps:{contains} ?idraw.
                                    ?statement pq:{object has role} ?role.
                                    FILTER (?role = wd:{output})
                                    BIND(CONCAT("mardi:", STRAFTER(STR(?idraw), STR(wd:))) AS ?ID)
                                    OPTIONAL {{ ?idraw rdfs:label ?Nameraw .
                                                FILTER (lang(?Nameraw) = 'en') }}
                                    BIND(COALESCE(?Nameraw, "No Label Provided!") AS ?Name)
                                    BIND("dependent" AS ?Type)
                                }}

                                # Filter results to only show defining statements that match the current quantity label
                                OPTIONAL {{
                                    ?task wdt:{contains} ?formulation.
                                    ?formulation wdt:{instance of} wd:{mathematical expression}.
                                    ?formulation p:{in defining formula} ?statement2.
                                    ?statement2 ps:{in defining formula} ?Symbol.
                                    ?statement2 pq:{symbol represents} ?quantityraw.
                                    BIND(CONCAT("mardi:", STRAFTER(STR(?quantityraw), STR(wd:))) AS ?quantity)
                                }}
                                
                                FILTER(?ID = ?quantity)       

                            # Initialize the Unit variable as empty
                            BIND("" AS ?Unit)
                            }}
                            ORDER BY ?label ?Type''',

                'parameters': '''SELECT ?Name ?Unit ?Symbol ?label

                                WHERE {{ 

                                VALUES ?task {{{0}}}

                                OPTIONAL {{ ?task rdfs:label ?labelraw .
                                            FILTER (lang(?labelraw) = 'en') }}
                                BIND(COALESCE(?labelraw, "No Label Provided!") AS ?label)

                                # Define type based on connection type (Input or Output)
                                
                                    ?task p:{contains} ?statement.
                                    ?statement ps:{contains} ?idraw.
                                    ?statement pq:{object has role} ?role.
                                    FILTER (?role = wd:{parameter})
                                    BIND(CONCAT("mardi:", STRAFTER(STR(?idraw), STR(wd:))) AS ?ID)
                                    OPTIONAL {{ ?idraw rdfs:label ?Nameraw .
                                                FILTER (lang(?Nameraw) = 'en') }}
                                    BIND(COALESCE(?Nameraw, "No Label Provided!") AS ?Name)
                                
                                
                                # Filter results to only show defining statements that match the current quantity label
                                OPTIONAL {{
                                    ?task wdt:{contains} ?formulation.
                                    ?formulation wdt:{instance of} wd:{mathematical expression}.
                                    ?formulation p:{in defining formula} ?statement2.
                                    ?statement2 ps:{in defining formula} ?Symbol.
                                    ?statement2 pq:{symbol represents} ?quantityraw.
                                    BIND(CONCAT("mardi:", STRAFTER(STR(?quantityraw), STR(wd:))) AS ?quantity)
                                }}
                                
                                FILTER(?ID = ?quantity)       

                            # Initialize the Unit variable as empty
                            BIND("" AS ?Unit)
                            }}
                            ORDER BY ?label'''
}


'''Module containing Providers for the Algorithm Documentation'''

from rdmo.options.providers import Provider

from ..helpers import define_setup
from ..queries import query_sources, query_sources_with_user_additions

class Algorithm(Provider):
    '''Algorithm Provider (MaRDI Portal / Wikidata / MathAlgoDB),
       No User Creation, Refresh Upon Selection
    '''

    search = True
    refresh = True

    def get_options(self, project, search=None, user=None, site=None):
        '''Queries MathModDB for user input'''
        if not search or len(search) < 3:
            return []

        # Define the sources to query
        query_id = 'algorithm/queries/provider_algorithm.sparql'
        sources = ['mathalgodb','mardi','wikidata']

        return query_sources(search, query_id, sources)

class RelatedAlgorithmWithoutCreation(Provider):
    '''Algorithm Provider (MaRDI Portal / Wikidata / MathAlgoDB),
       No User Creation, No Refresh Upon Selection
    '''

    search = True

    def get_options(self, project, search=None, user=None, site=None):

        if not search or len(search) < 3:
            return []

        # Define the query_setup
        setup = define_setup(
            query_attributes = ['algorithm'],
            query_id = 'algorithm/queries/provider_algorithm.sparql',
            sources = ['mathalgodb']
        )

        return query_sources_with_user_additions(
            search = search,
            project = project,
            setup = setup
        )

class AlgorithmicProblem(Provider):
    '''Algorithmic Problem Provider (MaRDI Portal / Wikidata / MathAlgoDB),
       No User Creation, Refresh Upon Selection
    '''

    search = True
    refresh = True

    def get_options(self, project, search=None, user=None, site=None):
        '''Queries MathModDB for user input'''
        if not search or len(search) < 3:
            return []

        # Define the sources to query
        query_id = 'algorithm/queries/provider_problem.sparql'
        sources = ['mathalgodb','mardi','wikidata']

        return query_sources(search, query_id, sources)

class RelatedAlgorithmicProblemWithCreation(Provider):
    '''Algorithmic Problem Provider (MaRDI Portal / Wikidata / MathAlgoDB),
       User Creation, No Refresh Upon Selection
    '''

    search = True

    def get_options(self, project, search=None, user=None, site=None):

        if not search or len(search) < 3:
            return []

        # Define the query_setup
        setup = define_setup(
            creation = True,
            query_attributes = ['problem'],
            query_id = 'algorithm/queries/provider_problem.sparql',
            sources = ['mathalgodb','mardi','wikidata']
        )

        return query_sources_with_user_additions(
            search = search,
            project = project,
            setup = setup
        )

class RelatedAlgorithmicProblemWithoutCreation(Provider):
    '''Algorithmic Problem Provider (MaRDI Portal / Wikidata / MathAlgoDB),
       No User Creation, No Refresh Upon Selection
    '''

    search = True

    def get_options(self, project, search=None, user=None, site=None):

        if not search or len(search) < 3:
            return []

        setup = define_setup(
            query_attributes = ['problem'],
            query_id = 'algorithm/queries/provider_problem.sparql',
            sources = ['mathalgodb']
        )

        return query_sources_with_user_additions(
            search = search,
            project = project,
            setup = setup
        )

class Software(Provider):
    '''Software Provider (MaRDI Portal / Wikidata / MathAlgoDB),
       No User Creation, Refresh Upon Selection
    '''

    search = True
    refresh = True

    def get_options(self, project, search=None, user=None, site=None):
        '''Queries MathModDB for user input'''
        if not search or len(search) < 3:
            return []

        # Define the sources to query
        query_id = 'algorithm/queries/provider_software.sparql'
        sources = ['mathalgodb','mardi','wikidata']

        return query_sources(search, query_id, sources)

class RelatedSoftwareWithCreation(Provider):
    '''Software Provider (MaRDI Portal / Wikidata / MathAlgoDB),
       User Creation, No Refresh Upon Selection
    '''

    search = True

    def get_options(self, project, search=None, user=None, site=None):

        if not search or len(search) < 3:
            return []

        # Define the query_setup
        setup = define_setup(
            creation = True,
            query_attributes = ['software'],
            query_id = 'algorithm/queries/provider_software.sparql',
            sources = ['mathalgodb','mardi','wikidata']
        )

        return query_sources_with_user_additions(
            search = search,
            project = project,
            setup = setup
        )

class Benchmark(Provider):
    '''Benchmark Provider (MaRDI Portal / Wikidata / MathAlgoDB),
       No User Creation, Refresh Upon Selection
    '''

    search = True
    refresh = True

    def get_options(self, project, search=None, user=None, site=None):
        '''Queries MathModDB for user input'''
        if not search or len(search) < 3:
            return []

        # Define the sources to query
        query_id = 'algorithm/queries/provider_benchmark.sparql'
        sources = ['mathalgodb','mardi','wikidata']

        return query_sources(search, query_id, sources)

class RelatedBenchmarkWithCreation(Provider):
    '''Benchmark Provider (MaRDI Portal / Wikidata / MathAlgoDB),
       User Creation, No Refresh Upon Selection
    '''

    search = True

    def get_options(self, project, search=None, user=None, site=None):

        if not search or len(search) < 3:
            return []

        # Define the query_setup
        setup = define_setup(
            creation = True,
            query_attributes = ['benchmark'],
            query_id = 'algorithm/queries/provider_benchmark.sparql',
            sources = ['mathalgodb','mardi','wikidata']
        )

        return query_sources_with_user_additions(
            search = search,
            project = project,
            setup = setup
        )



'''Module containing Models for the Model Documentation'''

from dataclasses import dataclass, field
from typing import Optional

from .constants import data_properties_per_class

from ..getters import get_items, get_mathmoddb, get_options
from ..helpers import split_value
from ..models import Relatant, RelatantWithClass

@dataclass
class RelatantWithQualifier:
    '''Data Class For Relatant Items With Qualifier'''
    id: Optional[str]
    label: Optional[str]
    description: Optional[str]
    qualifier: Optional[str]
    order: Optional[str]

    @classmethod
    def from_query(cls, raw: str) -> 'RelatantWithQualifier':
        '''Generate Class Item From Query'''
        if ">|<" in raw:
            raw, order = raw.split(" >|< ")
        else:
            order = None
        identifier, label, description, qualifier = raw.split(" | ", 3)
        return cls(
            id = identifier,
            label = label,
            description = description,
            qualifier = qualifier,
            order = order
        )

@dataclass
class ResearchField:
    '''Data Class For Research Field Item'''
    description_long: list[str] = field(default_factory=list)
    specialized_by: list[Relatant] = field(default_factory=list)
    specializes: list[Relatant] = field(default_factory=list)
    similar_to: list[Relatant] = field(default_factory=list)
    publications: list[Relatant] = field(default_factory=list)

    @classmethod
    def from_query(cls, raw_data: dict) -> 'ResearchField':
        '''Generate Class Item From Query'''

        data = raw_data[0]

        research_field = {
            # Get Long Description(s)
            'description_long': split_value(
                data = data,
                key = 'description_long'
            ),
            # Get Specialized By Relation(s)
            'specialized_by': split_value(
                data = data,
                key = 'specialized_by',
                transform = Relatant.from_query
            ),
            # Get Specializes Relation(s)
            'specializes': split_value(
                data = data,
                key = 'specializes',
                transform = Relatant.from_query
            ),
            # Get Similar To Relation(s)
            'similar_to': split_value(
                data = data,
                key = 'similar_to',
                transform = Relatant.from_query
            ),
            # Get Publication(s)
            'publications': split_value(
                data = data,
                key = 'publication',
                transform = Relatant.from_query
            )
        }

        return cls(
            **research_field
        )

@dataclass
class ResearchProblem:
    '''Data Class For Research Problem Item'''
    description_long: list[str] = field(default_factory=list)
    contained_in_field: list[Relatant] = field(default_factory=list)
    specialized_by: list[Relatant] = field(default_factory=list)
    specializes: list[Relatant] = field(default_factory=list)
    similar_to: list[Relatant] = field(default_factory=list)
    publications: list[Relatant] = field(default_factory=list)

    @classmethod
    def from_query(cls, raw_data: dict) -> 'ResearchProblem':
        '''Generate Class Item From Query'''

        data = raw_data[0]

        research_problem = {
            # Get Long Description(s)
            'description_long': split_value(
                data = data,
                key = 'description_long'
            ),
            # Get Field Relation(s)
            'contained_in_field': split_value(
                data = data,
                key = 'contained_in_field',
                transform = Relatant.from_query
            ),
            # Get Specialized By Relation(s)
            'specialized_by': split_value(
                data = data,
                key = 'specialized_by',
                transform = Relatant.from_query
            ),
            # Get Specializes Relation(s)
            'specializes': split_value(
                data = data,
                key = 'specializes',
                transform = Relatant.from_query
            ),
            # Get Similar To Relation(s)
            'similar_to': split_value(
                data = data,
                key = 'similar_to',
                transform = Relatant.from_query
            ),
            # Get Publication(s)
            'publications': split_value(
                data = data,
                key = 'publication',
                transform = Relatant.from_query
            )
        }

        return cls(
            **research_problem
        )

@dataclass
class MathematicalModel:
    '''Data Class For Mathematical Model Item'''
    description_long: list[str] = field(default_factory=list)
    properties: dict[int, str] = field(default_factory=dict)
    models: list[Relatant] = field(default_factory=list)
    assumes: list[Relatant] = field(default_factory=list)
    contains_formulation: list[Relatant] = field(default_factory=list)
    contains_boundary_condition: list[Relatant] = field(default_factory=list)
    contains_constraint_condition: list[Relatant] = field(default_factory=list)
    contains_coupling_condition: list[Relatant] = field(default_factory=list)
    contains_initial_condition: list[Relatant] = field(default_factory=list)
    contains_final_condition: list[Relatant] = field(default_factory=list)
    used_by: list[Relatant] = field(default_factory=list)
    specializes: list[Relatant] = field(default_factory=list)
    specialized_by: list[Relatant] = field(default_factory=list)
    approximates: list[Relatant] = field(default_factory=list)
    approximated_by: list[Relatant] = field(default_factory=list)
    contains_model: list[Relatant] = field(default_factory=list)
    contained_in_model: list[Relatant] = field(default_factory=list)
    discretized_by: list[Relatant] = field(default_factory=list)
    discretizes: list[Relatant] = field(default_factory=list)
    linearized_by: list[Relatant] = field(default_factory=list)
    linearizes: list[Relatant] = field(default_factory=list)
    similar_to: list[Relatant] = field(default_factory=list)
    publications: list[Relatant] = field(default_factory=list)

    @classmethod
    def from_query(cls, raw_data: dict) -> 'MathematicalModel':
        '''Generate Class Item From Query'''

        mathmoddb = get_mathmoddb()
        items = get_items()

        data = raw_data[0]

        mathematical_model = {
            # Get Long Description(s)
            'description_long': split_value(
                data = data,
                key = 'description_long'
            ),
            # Get Properties
            'properties': {
                idx: [mathmoddb[prop]]
                for idx, prop in enumerate(data_properties_per_class['model'])
                if data.get(prop, {}).get('value') == 'True'
            },
            # Get Problem Relation(s)
            'models': split_value(
                data = data,
                key = 'models',
                transform = Relatant.from_query
            ),
            # Get Assumption Relation(s)
            'assumes': split_value(
                data = data,
                key = 'assumes',
                transform = RelatantWithQualifier.from_query
            ),
            # Get Contains Formulation Relation(s)
            'contains_formulation': split_value(
                data = data,
                key = 'contains_formulation',
                transform = RelatantWithQualifier.from_query,
                object_role = lambda item: item.qualifier == ''
            ),
            # Get Contains Boundary Condition Relation(s)
            'contains_boundary_condition': split_value(
                data = data,
                key = 'contains_formulation',
                transform = RelatantWithQualifier.from_query,
                object_role = lambda item: item.qualifier == f'mardi:{items["boundary condition"]}'
            ),
            # Get Contains Constraint Condition Relation(s)
            'contains_constraint_condition': split_value(
                data = data,
                key = 'contains_formulation',
                transform = RelatantWithQualifier.from_query,
                object_role = lambda item: item.qualifier == f'mardi:{items["constraint"]}'
            ),
            # Get Contains Coupling Condition Relation(s)
            'contains_coupling_condition': split_value(
                data = data,
                key = 'contains_formulation',
                transform = RelatantWithQualifier.from_query,
                object_role = lambda item: item.qualifier == f'mardi:{items["coupling condition"]}'
            ),
            # Get Contains Initial Condition Relation(s)
            'contains_initial_condition': split_value(
                data = data,
                key = 'contains_formulation',
                transform = RelatantWithQualifier.from_query,
                object_role = lambda item: item.qualifier == f'mardi:{items["initial condition"]}'
            ),
            # Get Contains Final Condition Relation(s)
            'contains_final_condition': split_value(
                data = data,
                key = 'contains_formulation',
                transform = RelatantWithQualifier.from_query,
                object_role = lambda item: item.qualifier == f'mardi:{items["final condition"]}'
            ),
            # Get Task Relation(s)
            'used_by': split_value(
                data = data,
                key = 'used_by',
                transform = Relatant.from_query
            ),
            # Get Specialized By Relation(s)
            'specialized_by': split_value(
                data = data,
                key = 'specialized_by',
                transform = RelatantWithQualifier.from_query
            ),
            # Get Specializes Relation(s)
            'specializes': split_value(
                data = data,
                key = 'specializes',
                transform = RelatantWithQualifier.from_query
            ),
            # Get Approximated By Relation(s)
            'approximated_by': split_value(
                data = data,
                key = 'approximated_by',
                transform = Relatant.from_query
            ),
            # Get Approximates Relation(s)
            'approximates': split_value(
                data = data,
                key = 'approximates',
                transform = Relatant.from_query
            ),
            # Get Contained In Model Relation(s)
            'contained_in_model': split_value(
                data = data,
                key = 'contained_in_model',
                transform = Relatant.from_query
            ),
            # Get Contains Model Relation(s)
            'contains_model': split_value(
                data = data,
                key = 'contains_model',
                transform = Relatant.from_query
            ),
            # Get Discretized By Relation(s)
            'discretized_by': split_value(
                data = data,
                key = 'discretized_by',
                transform = Relatant.from_query
            ),
            # Get Discretizes Relation(s)
            'discretizes': split_value(
                data = data,
                key = 'discretizes',
                transform = Relatant.from_query
            ),
            # Get Linearized By Relation(s)
            'linearized_by': split_value(
                data = data,
                key = 'linearized_by',
                transform = Relatant.from_query
            ),
            # Get Linearizes Relation(s)
            'linearizes': split_value(
                data = data,
                key = 'linearizes',
                transform = Relatant.from_query
            ),
            # Get Similar To Relation(s)
            'similar_to': split_value(
                data = data,
                key = 'similar_to',
                transform = Relatant.from_query
            ),
            # Get Publication(s)
            'publications': split_value(
                data = data,
                key = 'publication',
                transform = Relatant.from_query
            )
        }

        return cls(
            **mathematical_model
        )

@dataclass
class QuantityOrQuantityKind:
    '''Data Class For Quantity [Kind] Item'''
    description_long: list[str] = field(default_factory=list)
    reference: dict[int, list[str]] = field(default_factory=dict)
    qclass: Optional[str] = None
    properties: dict[int, str] = field(default_factory=dict)
    formulas: list[str] = field(default_factory=list)
    symbols: list[str] = field(default_factory=list)
    contains_quantity: list[Relatant] = field(default_factory=list)
    specialized_by: list[Relatant] = field(default_factory=list)
    specializes: list[Relatant] = field(default_factory=list)
    approximated_by: list[Relatant] = field(default_factory=list)
    approximates: list[Relatant] = field(default_factory=list)
    discretized_by: list[Relatant] = field(default_factory=list)
    discretizes: list[Relatant] = field(default_factory=list)
    linearized_by: list[Relatant] = field(default_factory=list)
    linearizes: list[Relatant] = field(default_factory=list)
    nondimensionalized_by: list[Relatant] = field(default_factory=list)
    nondimensionalizes: list[Relatant] = field(default_factory=list)
    similar_to: list[Relatant] = field(default_factory=list)
    publications: list[Relatant] = field(default_factory=list)

    @classmethod
    def from_query(cls, raw_data: dict) -> 'QuantityOrQuantityKind':
        '''Generate Class Item From Query'''

        mathmoddb = get_mathmoddb()
        options = get_options()

        data = raw_data[0]

        quantity = {
            # Get Long Description(s)
            'description_long': split_value(
                data = data,
                key = 'description_long'
            ),
            # Get Reference
            'reference': {
                idx: [options['QUDT'], data[prop]['value'].removeprefix('https://qudt.org/vocab/')]
                for idx, prop in enumerate(['reference'])
                if data.get(prop, {}).get('value')
            },
            # Get Quantity or Quantity Kind Class
            'qclass': data.get('class', {}).get('value'),
            # Get Properties
            'properties': {
                idx: [mathmoddb[prop]]
                for idx, prop in enumerate(data_properties_per_class['quantity'])
                if data.get(prop, {}).get('value') == 'True'
            },
            # Get Formulas
            'formulas': split_value(
                data = data,
                key = 'formulas'
            ),
            # Get Symbols
            'symbols': split_value(
                data = data,
                key = 'contains_quantity',
                transform = lambda s: s.split(' | ', 1)[0]
            ),
            # Get contained Quantities
            'contains_quantity': split_value(
                data = data,
                key = 'contains_quantity',
                transform = lambda q: Relatant.from_query(q.split(' | ', 1)[1])
            ),
            # Get Specialized By Relation(s)
            'specialized_by': split_value(
                data = data,
                key = 'specialized_by',
                transform = RelatantWithClass.from_query
            ),
            # Get Specializes Relation(s)
            'specializes': split_value(
                data = data,
                key = 'specializes',
                transform = RelatantWithClass.from_query
            ),
            # Get Approximated By Relation(s)
            'approximated_by': split_value(
                data = data,
                key = 'approximated_by',
                transform = RelatantWithClass.from_query
            ),
            # Get Approximates Relation(s)
            'approximates': split_value(
                data = data,
                key = 'approximates',
                transform = RelatantWithClass.from_query
            ),
            # Get Discretized By Relation(s)
            'discretized_by': split_value(
                data = data,
                key = 'discretized_by',
                transform = RelatantWithClass.from_query
            ),
            # Get Discretizes Relation(s)
            'discretizes': split_value(
                data = data,
                key = 'discretizes',
                transform = RelatantWithClass.from_query
            ),
            # Get Linearized By Relation(s)
            'linearized_by': split_value(
                data = data,
                key = 'linearized_by',
                transform = RelatantWithClass.from_query
            ),
            # Get Linearizes Relation(s)
            'linearizes': split_value(
                data = data,
                key = 'linearizes',
                transform = RelatantWithClass.from_query
            ),
            # Get Nondimesionalized By Relation(s)
            'nondimensionalized_by': split_value(
                data = data,
                key = 'nondimensionalized_by',
                transform = RelatantWithClass.from_query
            ),
            # Get Nondimensionalizes Relation(s)
            'nondimensionalizes': split_value(
                data = data,
                key = 'nondimensionalizes',
                transform = RelatantWithClass.from_query
            ),
            # Get Similar To Relation(s)
            'similar_to': split_value(
                data = data,
                key = 'similar_to',
                transform = RelatantWithClass.from_query
            ),
            # Get Publication(s)
            'publications': split_value(
                data = data,
                key = 'publication',
                transform = Relatant.from_query
            )
        }

        return cls(
            **quantity
        )

@dataclass
class MathematicalFormulation:
    '''Data Class For Formulation Item'''
    description_long: list[str] = field(default_factory=list)
    properties: dict[int, str] = field(default_factory=dict)
    formulas: list[str] = field(default_factory=list)
    symbols: list[str] = field(default_factory=list)
    contains_quantity: list[Relatant] = field(default_factory=list)
    assumes: list[Relatant] = field(default_factory=list)
    contains_formulation: list[Relatant] = field(default_factory=list)
    contains_assumption: list[Relatant] = field(default_factory=list)
    contains_boundary_condition: list[Relatant] = field(default_factory=list)
    contains_final_condition: list[Relatant] = field(default_factory=list)
    contains_initial_condition: list[Relatant] = field(default_factory=list)
    contains_constraint_condition: list[Relatant] = field(default_factory=list)
    contains_coupling_condition: list[Relatant] = field(default_factory=list)
    specialized_by: list[Relatant] = field(default_factory=list)
    specializes: list[Relatant] = field(default_factory=list)
    discretized_by: list[Relatant] = field(default_factory=list)
    discretizes: list[Relatant] = field(default_factory=list)
    approximated_by: list[Relatant] = field(default_factory=list)
    approximates: list[Relatant] = field(default_factory=list)
    linearized_by: list[Relatant] = field(default_factory=list)
    linearizes: list[Relatant] = field(default_factory=list)
    nondimensionalized_by: list[Relatant] = field(default_factory=list)
    nondimensionalizes: list[Relatant] = field(default_factory=list)
    similar_to: list[Relatant] = field(default_factory=list)
    publications: list[Relatant] = field(default_factory=list)


    @classmethod
    def from_query(cls, raw_data: dict) -> 'MathematicalFormulation':
        '''Generate Class Item From Query'''

        mathmoddb = get_mathmoddb()
        items = get_items()

        data = raw_data[0]

        mathematical_formulation = {
            # Get Long Description(s)
            'description_long': split_value(
                data = data,
                key = 'description_long'
            ),
            # Get Properties
            'properties': {
                idx: [mathmoddb[prop]]
                for idx, prop in enumerate(data_properties_per_class['formulation'])
                if data.get(prop, {}).get('value') == 'True'
            },
            # Get Formulas
            'formulas': split_value(
                data = data,
                key = 'formulas'
            ),
            # Get Symbols
            'symbols': split_value(
                data = data,
                key = 'contains_quantity',
                transform = lambda s: s.split(' | ', 1)[0]
            ),
            # Get contained Quantities
            'contains_quantity': split_value(
                data = data,
                key = 'contains_quantity',
                transform = lambda q: Relatant.from_query(q.split(' | ', 1)[1])
            ),
            # Get Assumption Relation(s)
            'assumes': split_value(
                data = data,
                key = 'assumes',
                transform = Relatant.from_query
            ),
            # Get Contains Formulation Relation(s)
            'contains_formulation': split_value(
                data = data,
                key = 'contains_formulation',
                transform = RelatantWithQualifier.from_query,
                object_role = lambda item: item.qualifier == ''
            ),
            # Get Contains Boundary Condition Relation(s)
            'contains_boundary_condition': split_value(
                data = data,
                key = 'contains_formulation',
                transform = RelatantWithQualifier.from_query,
                object_role = lambda item: item.qualifier == f'mardi:{items["boundary condition"]}'
            ),
            # Get Contains Constraint Condition Relation(s)
            'contains_constraint_condition': split_value(
                data = data,
                key = 'contains_formulation',
                transform = RelatantWithQualifier.from_query,
                object_role = lambda item: item.qualifier == f'mardi:{items["constraint"]}'
            ),
            # Get Contains Coupling Condition Relation(s)
            'contains_coupling_condition': split_value(
                data = data,
                key = 'contains_formulation',
                transform = RelatantWithQualifier.from_query,
                object_role = lambda item: item.qualifier == f'mardi:{items["coupling condition"]}'
            ),
            # Get Contains Initial Condition Relation(s)
            'contains_initial_condition': split_value(
                data = data,
                key = 'contains_formulation',
                transform = RelatantWithQualifier.from_query,
                object_role = lambda item: item.qualifier == f'mardi:{items["initial condition"]}'
            ),
            # Get Contains Final Condition Relation(s)
            'contains_final_condition': split_value(
                data = data,
                key = 'contains_formulation',
                transform = RelatantWithQualifier.from_query,
                object_role = lambda item: item.qualifier == f'mardi:{items["final condition"]}'
            ),
            # Get Specialized By Relation(s)
            'specialized_by': split_value(
                data = data,
                key = 'specialized_by',
                transform = RelatantWithQualifier.from_query
            ),
            # Get Specializes Relation(s)
            'specializes': split_value(
                data = data,
                key = 'specializes',
                transform = RelatantWithQualifier.from_query
            ),
            # Get Approximated By Relation(s)
            'approximated_by': split_value(
                data = data,
                key = 'approximated_by',
                transform = Relatant.from_query
            ),
            # Get Approximates Relation(s)
            'approximates': split_value(
                data = data,
                key = 'approximates',
                transform = Relatant.from_query
            ),
            # Get Discretized By Relation(s)
            'discretized_by': split_value(
                data = data,
                key = 'discretized_by',
                transform = Relatant.from_query
            ),
            # Get Discretizes Relation(s)
            'discretizes': split_value(
                data = data,
                key = 'discretizes',
                transform = Relatant.from_query
            ),
            # Get Linearized By Relation(s)
            'linearized_by': split_value(
                data = data,
                key = 'linearized_by',
                transform = Relatant.from_query
            ),
            # Get Linearizes Relation(s)
            'linearizes': split_value(
                data = data,
                key = 'linearizes',
                transform = Relatant.from_query
            ),
            # Get Nondimensionalized By Relation(s)
            'nondimensionalized_by': split_value(
                data = data,
                key = 'nondimensionalized_by',
                transform = Relatant.from_query
            ),
            # Get Nondimesionalizes Relation(s)
            'nondimensionalizes': split_value(
                data = data,
                key = 'nondimensionalizes',
                transform = Relatant.from_query
            ),
            # Get Similar To Relation(s)
            'similar_to': split_value(
                data = data,
                key = 'similar_to',
                transform = Relatant.from_query
            ),
            # Get Publication(s)
            'publications': split_value(
                data = data,
                key = 'publication',
                transform = Relatant.from_query
            )
        }

        return cls(
            **mathematical_formulation
        )

@dataclass
class Task:
    '''Data Class For Task Item'''
    description_long: list[str] = field(default_factory=list)
    properties: dict[int, str] = field(default_factory=dict)
    assumes: list[Relatant] = field(default_factory=list)
    contains_formulation: list[Relatant] = field(default_factory=list)
    contains_boundary_condition: list[Relatant] = field(default_factory=list)
    contains_final_condition: list[Relatant] = field(default_factory=list)
    contains_initial_condition: list[Relatant] = field(default_factory=list)
    contains_constraint_condition: list[Relatant] = field(default_factory=list)
    contains_coupling_condition: list[Relatant] = field(default_factory=list)
    contains_input: list[Relatant] = field(default_factory=list)
    contains_output: list[Relatant] = field(default_factory=list)
    contains_objective: list[Relatant] = field(default_factory=list)
    contains_parameter: list[Relatant] = field(default_factory=list)
    contains_constant: list[Relatant] = field(default_factory=list)
    specializes: list[Relatant] = field(default_factory=list)
    specialized_by: list[Relatant] = field(default_factory=list)
    approximates: list[Relatant] = field(default_factory=list)
    approximated_by: list[Relatant] = field(default_factory=list)
    contains_task: list[Relatant] = field(default_factory=list)
    contained_in_task: list[Relatant] = field(default_factory=list)
    discretized_by: list[Relatant] = field(default_factory=list)
    discretizes: list[Relatant] = field(default_factory=list)
    linearized_by: list[Relatant] = field(default_factory=list)
    linearizes: list[Relatant] = field(default_factory=list)
    similar_to: list[Relatant] = field(default_factory=list)
    publications: list[Relatant] = field(default_factory=list)

    @classmethod
    def from_query(cls, raw_data: dict) -> 'Task':
        '''Generate Class Item From Query'''

        mathmoddb = get_mathmoddb()
        items = get_items()

        data = raw_data[0]

        task = {
            # Get Long Description(s)
            'description_long': split_value(
                data = data,
                key = 'description_long'
            ),
            # Get Properties
            'properties': {
                idx: [mathmoddb[prop]]
                for idx, prop in enumerate(data_properties_per_class['task'])
                if data.get(prop, {}).get('value') == 'True'
            },
            # Get Assumption Relation(s)
            'assumes': split_value(
                data = data,
                key = 'assumes',
                transform = Relatant.from_query
            ),
            # Get Contains Formulation Relation(s)
            'contains_formulation': split_value(
                data = data,
                key = 'contains_formulation',
                transform = RelatantWithQualifier.from_query,
                object_role = lambda item: item.qualifier == ''
            ),
            # Get Contains Boundary Condition Relation(s)
            'contains_boundary_condition': split_value(
                data = data,
                key = 'contains_formulation',
                transform = RelatantWithQualifier.from_query,
                object_role = lambda item: item.qualifier == f'mardi:{items["boundary condition"]}'
            ),
            # Get Contains Constraint Condition Relation(s)
            'contains_constraint_condition': split_value(
                data = data,
                key = 'contains_formulation',
                transform = RelatantWithQualifier.from_query,
                object_role = lambda item: item.qualifier == f'mardi:{items["constraint"]}'
            ),
            # Get Contains Coupling Condition Relation(s)
            'contains_coupling_condition': split_value(
                data = data,
                key = 'contains_formulation',
                transform = RelatantWithQualifier.from_query,
                object_role = lambda item: item.qualifier == f'mardi:{items["coupling condition"]}'
            ),
            # Get Contains Initial Condition Relation(s)
            'contains_initial_condition': split_value(
                data = data,
                key = 'contains_formulation',
                transform = RelatantWithQualifier.from_query,
                object_role = lambda item: item.qualifier == f'mardi:{items["initial condition"]}'
            ),
            # Get Contains Final Condition Relation(s)
            'contains_final_condition': split_value(
                data = data,
                key = 'contains_formulation',
                transform = RelatantWithQualifier.from_query,
                object_role = lambda item: item.qualifier == f'mardi:{items["final condition"]}'
            ),
            # Get Contains Input Relation(s)
            'contains_input': split_value(
                data = data,
                key = 'contains_quantity',
                transform = RelatantWithQualifier.from_query,
                object_role = lambda item: item.qualifier == f'mardi:{items["input"]}'
            ),
            # Get Contains Output Relation(s)
            'contains_output': split_value(
                data = data,
                key = 'contains_quantity',
                transform = RelatantWithQualifier.from_query,
                object_role = lambda item: item.qualifier == f'mardi:{items["output"]}'
            ),
            # Get Contains Objective Relation(s)
            'contains_objective': split_value(
                data = data,
                key = 'contains_quantity',
                transform = RelatantWithQualifier.from_query,
                object_role = lambda item: item.qualifier == f'mardi:{items["objective function"]}'
            ),
            # Get Contains Parameter Relation(s)
            'contains_parameter': split_value(
                data = data,
                key = 'contains_quantity',
                transform = RelatantWithQualifier.from_query,
                object_role = lambda item: item.qualifier == f'mardi:{items["parameter"]}'
            ),
            # Get Contains Constant Relation(s)
            'contains_constant': split_value(
                data = data,
                key = 'contains_quantity',
                transform = RelatantWithQualifier.from_query,
                object_role = lambda item: item.qualifier == f'mardi:{items["constant"]}'
            ),
            # Get Specialized By Relation(s)
            'specialized_by': split_value(
                data = data,
                key = 'specialized_by',
                transform = RelatantWithQualifier.from_query
            ),
            # Get Specializes Relation(s)
            'specializes': split_value(
                data = data,
                key = 'specializes',
                transform = RelatantWithQualifier.from_query
            ),
            # Get Approximated By Relation(s)
            'approximated_by': split_value(
                data = data,
                key = 'approximated_by',
                transform = Relatant.from_query
            ),
            # Get Approximates Relation(s)
            'approximates': split_value(
                data = data,
                key = 'approximates',
                transform = Relatant.from_query
            ),
            # Get Contained In Relation(s)
            'contained_in_task': split_value(
                data = data,
                key = 'contained_in_task',
                transform = RelatantWithQualifier.from_query
            ),
            # Get Contains Relation(s)
            'contains_task': split_value(
                data = data,
                key = 'contains_task',
                transform = RelatantWithQualifier.from_query
            ),
            # Get Discretized By Relation(s)
            'discretized_by': split_value(
                data = data,
                key = 'discretized_by',
                transform = Relatant.from_query
            ),
            # Get Discretizes Relation(s)
            'discretizes': split_value(
                data = data,
                key = 'discretizes',
                transform = Relatant.from_query
            ),
            # Get Linearized By Relation(s)
            'linearized_by': split_value(
                data = data,
                key = 'linearized_by',
                transform = Relatant.from_query
            ),
            # Get Linearizes Relation(s)
            'linearizes': split_value(
                data = data,
                key = 'linearizes',
                transform = Relatant.from_query
            ),
            # Get Similar To Relation(s)
            'similar_to': split_value(
                data = data,
                key = 'similar_to',
                transform = Relatant.from_query
            ),
            # Get Publication(s)
            'publications': split_value(
                data = data,
                key = 'publication',
                transform = Relatant.from_query
            )
        }

        return cls(
            **task
            )



'''Module containing Checks for the Model Documentation'''

from rdmo.domain.models import Attribute

from .constants import data_properties_check, section_map

from ..config import BASE_URI
from ..getters import get_mathmoddb

class Checks:
    '''Check Class checks User Answers upon Transfer to MaRDI Portal'''
    def __init__(self):
        self.mathmoddb = get_mathmoddb()
        self.err = []

    def error_message(self, section, page, message):
        '''Generate Error Message'''
        return f"{section} (Page {page}): {message}"

    def id_name_description(self, project, data):
        '''Perform ID, Name and Description Checks:
            - ID, Name, Description present'''

        okeys = ('model', 'formulation', 'quantity', 'task', 'problem', 'field', 'publication')

        for okey, ovalue in data.items():
            if okey in okeys:
                values = project.values.filter(
                    snapshot=None,
                    attribute=Attribute.objects.get(
                        uri=f'{BASE_URI}domain/{okey}'
                    )
                )
                for ikey, ivalue in ovalue.items():
                    page_name = values.get(set_index=ikey).text
                    if not ivalue.get('ID'):
                        self.err.append(
                            self.error_message(
                                section = section_map[okey],
                                page = page_name,
                                message = 'Missing ID'
                            )
                        )
                    if not ivalue.get('Name'):
                        self.err.append(
                            self.error_message(
                                section = section_map[okey],
                                page = page_name,
                                message = 'Missing Name'
                            )
                        )
                    if not ivalue.get('Description'):
                        self.err.append(
                            self.error_message(
                                section = section_map[okey],
                                page = page_name,
                                message = 'Missing Short Description'
                            )
                        )
                    if ivalue.get('Description') and len(ivalue['Description']) > 250:
                        self.err.append(
                            self.error_message(
                                section = section_map[okey],
                                page = page_name,
                                message = 'Short Description Too Long'
                            )
                        )

    def properties(self, project,data):
        '''Perform Property Checks:
            - Properties present
            - Properties consistent'''

        okeys = ('model', 'formulation', 'quantity', 'task')

        for okey, ovalue in data.items():
            if okey not in okeys:
                continue
            values = project.values.filter(
                snapshot=None,
                attribute=Attribute.objects.get(
                    uri = f'{BASE_URI}domain/{okey}'
                )
            )
            for ikey, ivalue in ovalue.items():
                page_name = values.get(set_index=ikey).text
                if not ivalue.get('Properties'):
                    self.err.append(
                        self.error_message(
                            section = section_map[okey],
                            page = page_name,
                            message = 'Missing Properties'
                        )
                    )
                else:
                    properties = ivalue['Properties'].values()
                    for key, value in data_properties_check.items():
                        if {self.mathmoddb[key[0]], self.mathmoddb[key[1]]}.issubset(properties):
                            self.err.append(
                                self.error_message(
                                    section = section_map[okey],
                                    page = page_name,
                                    message = f'Inconsistent Properties {value}'
                                )
                            )

    def model(self, project, data):
        '''Perform Model Checks:
            - Connections present
            - Relations present'''

        values = project.values.filter(
            snapshot = None,
            attribute = Attribute.objects.get(
                uri = f"{BASE_URI}domain/model"
            )
        )

        if not data.get('model'):
            self.err.append('Mathematical Model: No Mathematical Model Entered')
            return

        for ikey, ivalue in data.get('model',{}).items():
            page_name = values.get(set_index=ikey).text
            # Check Connections
            if not ivalue.get('RelationRP'):
                self.err.append(
                    self.error_message(
                        'Mathematical Model',
                        page_name,
                        'Missing Research Problem'
                    )
                )
            if not ivalue.get('RelationT'):
                self.err.append(
                    self.error_message(
                        'Mathematical Model',
                        page_name,
                        'Missing Computational Task'
                    )
                )
            if not ivalue.get('RelationMF'):
                self.err.append(
                    self.error_message(
                        'Mathematical Model',
                        page_name,
                        'Missing Mathematical Expression'
                    )
                )
            # Relation Connections
            if any(
                'MISSING RELATION TYPE' in val
                for val in ivalue.get('RelationMF', {}).values()
            ):
                self.err.append(
                    self.error_message(
                        'Mathematical Model',
                        page_name,
                        'Missing Relation Type (Mathematical Expression)'
                    )
                )
            if any(
                'MISSING RELATION TYPE' in val
                for val in ivalue.get('RelationMM', {}).values()
            ):
                self.err.append(
                    self.error_message(
                        'Mathematical Model',
                        page_name,
                        'Missing Relation Type (Mathematical Model)'
                    )
                )
            # Check Qualifier
            for mkey, mval in ivalue.get('RelationMM', {}).items():
                if mval[0] in (self.mathmoddb['specializes'], self.mathmoddb['specialized_by']):
                    if not ivalue.get('assumption', {}).get(mkey):
                        self.err.append(
                            self.error_message(
                                'Mathematical Model',
                                page_name,
                                'Missing Assumption '
                                '(Specializes / Specialized By Mathematical Model)'
                            )
                        )
            if ivalue.get('number'):
                if not len(ivalue['number']) == len(ivalue['RelationMF']):
                    self.err.append(
                        self.error_message(
                            'Mathematical Model',
                            page_name,
                            'Missing Order Number (Mathematical Expression)'
                        )
                    )
                numbers = set(map(int, ivalue["number"].values()))
                expected = set(range(1, len(ivalue["number"]) + 1))
                if numbers != expected:
                    self.err.append(
                        self.error_message(
                            'Mathematical Model',
                            page_name,
                            'Incorrect Order Number (Mathematical Expression)'
                        )
                    )

    def task(self, project, data):
        '''Perform Task Checks:
            - Connections present
            - Relations present
            - Qualifier present'''

        values = project.values.filter(
            snapshot = None,
            attribute = Attribute.objects.get(
                uri = f"{BASE_URI}domain/task"
            )
        )

        if not data.get('task'):
            self.err.append('Computational Task: No Computational Task Entered')
            return

        for ikey, ivalue in data.get('task',{}).items():
            page_name = values.get(set_index=ikey).text
            # Check Connections
            if not ivalue.get('RelationMF'):
                self.err.append(
                    self.error_message(
                        'Computational Task',
                        page_name,
                        'Missing Mathematical Expression'
                    )
                )
            if not ivalue.get('RelationQQK'):
                self.err.append(
                    self.error_message(
                        'Computational Task',
                        page_name,
                        'Missing Quantity / Quantity Kind'
                    )
                )
            # Relation Connections
            if any(
                'MISSING RELATION TYPE' in val
                for val in ivalue.get('RelationMF', {}).values()
            ):
                self.err.append(
                    self.error_message(
                        'Computational Task',
                        page_name,
                        'Missing Relation Type (Mathematical Expression)'
                    )
                )
            if any(
                'MISSING RELATION TYPE' in val
                for val in ivalue.get('RelationQQK', {}).values()
            ):
                self.err.append(
                    self.error_message(
                        'Computational Task',
                        page_name,
                        'Missing Relation Type (Quantity / Quantity Kind)'
                    )
                )
            if any(
                'MISSING RELATION TYPE' in val
                for val in ivalue.get('RelationT', {}).values()
            ):
                self.err.append(
                    self.error_message(
                        'Computational Task',
                        page_name,
                        'Missing Relation Type (Computational Task)'
                    )
                )
            # Check Qualifier
            for tkey, tval in ivalue.get('RelationT', {}).items():
                if tval[0] in (self.mathmoddb['specializes'], self.mathmoddb['specialized_by']):
                    if not ivalue.get('assumption', {}).get(tkey):
                        self.err.append(
                            self.error_message(
                                'Computational Task',
                                page_name,
                                'Missing Assumption'
                                '(Specializes / Specialized By Computational Task)'
                            )
                        )
            for tkey, tval in ivalue.get('RelationT', {}).items():
                if tval[0] in (self.mathmoddb['contains'], self.mathmoddb['contained_in']):
                    if not ivalue.get('task_number', {}).get(tkey):
                        self.err.append(
                            self.error_message(
                                'Computational Task',
                                page_name,
                                'Missing Order Number'
                                '(Conatins / Contained In Computational Task)'
                            )
                        )

    def formulation(self, project, data):
        '''Perform Formulation Checks:
            - Formula present
            - Element present
            - Relations present'''

        values = project.values.filter(
            snapshot = None,
            attribute = Attribute.objects.get(
                uri = f"{BASE_URI}domain/formulation"
            )
        )

        if not data.get('formulation'):
            self.err.append('Mathematical Expression: No Mathematical Expression Entered')
            return

        for ikey, ivalue in data.get('formulation',{}).items():
            page_name = values.get(set_index=ikey).text
            # Check Formula
            if not ivalue.get('Formula'):
                self.err.append(
                    self.error_message(
                        'Mathematical Expression',
                        page_name,
                        'Missing Mathematical Expression Formula'
                    )
                )
            # Check Element
            if not ivalue.get('element'):
                self.err.append(
                    self.error_message(
                        'Mathematical Expression',
                        page_name,
                        'Missing Mathematical Expression Element Information'
                    )
                )
            else:
                not_symbol = False
                not_quantity = False
                for evalue in ivalue['element'].values():
                    if not evalue.get('symbol'):
                        not_symbol = True
                    if not evalue.get('quantity'):
                        not_quantity = True
                if not_symbol:
                    self.err.append(
                        self.error_message(
                            'Mathematical Expression',
                            page_name,
                            'Missing Mathematical Expression Symbol'
                        )
                    )
                if not_quantity:
                    self.err.append(
                        self.error_message(
                            'Mathematical Expression',
                            page_name,
                            'Missing Mathematical Expression Quantity'
                        )
                    )
            # Relation Connections
            if any(
                'MISSING RELATION TYPE' in val
                for val in ivalue.get('RelationMF1', {}).values()
            ):
                self.err.append(
                    self.error_message(
                        'Mathematical Expression',
                        page_name,
                        'Missing Relation Type (Mathematical Expression I)'
                    )
                )
            if any(
                'MISSING RELATION TYPE' in val
                for val in ivalue.get('RelationMF2', {}).values()
            ):
                self.err.append(
                    self.error_message(
                        'Mathematical Expression',
                        page_name,
                        'Missing Relation Type (Mathematical Expression II)'
                    )
                )
            # Check Qualifier
            for mkey, mval in ivalue.get('RelationMF2', {}).items():
                if mval[0] in (self.mathmoddb['specializes'], self.mathmoddb['specialized_by']):
                    if not ivalue.get('assumption', {}).get(mkey):
                        self.err.append(
                            self.error_message(
                                'Mathematical Expression',
                                page_name,
                                'Missing Assumption '
                                '(Specializes / Specialized By Mathematical Expression)'
                            )
                        )

    def quantity(self, project, data):
        '''Perform Quantity Checks:
            - Class present
            - Formula is Definition 
            - Elements of Formula present
            - Relations present'''

        values = project.values.filter(
            snapshot = None,
            attribute = Attribute.objects.get(
                uri = f"{BASE_URI}domain/quantity"
            )
        )

        if not data.get('quantity'):
            self.err.append('Quantity [Kind]: No Quantity [Kind] Entered')
            return

        for ikey, ivalue in data.get('quantity',{}).items():
            page_name = values.get(set_index=ikey).text
            # Check Class
            if not ivalue.get('QorQK'):
                self.err.append(
                    self.error_message(
                        'Quantity [Kind]',
                        page_name,
                        'Missing [Kind] Class'
                    )
                )
            # Check Formula
            if ivalue.get('Formula'):
                # Check \equiv sign
                for formula in ivalue['Formula'].values():
                    if not any(
                        equiv in formula
                        for equiv in ('>≡</', '>&#x2261;</', '>&equiv;</', '\\equiv', '\\Equiv')
                    ):
                        self.err.append(
                            self.error_message(
                                'Quantity [Kind]',
                                page_name,
                                r'Inconsistent Quantity [Kind] Definition (missing \equiv)'
                            )
                        )
                # Check Element
                if not ivalue.get('element'):
                    self.err.append(
                        self.error_message(
                            'Quantity / Quantity Kind',
                            page_name,
                            'Missing Quantity [Kind] Definition Element Information'
                        )
                    )
                else:
                    not_symbol = False
                    not_quantity = False
                    for evalue in ivalue['element'].values():
                        if not evalue.get('symbol'):
                            not_symbol = True
                        if not evalue.get('quantity'):
                            not_quantity = True
                    if not_symbol:
                        self.err.append(
                            self.error_message(
                                'Quantity [Kind]',
                                page_name,
                                'Missing Quantity [Kind] Definition Symbol'
                            )
                        )
                    if not_quantity:
                        self.err.append(
                            self.error_message(
                                'Quantity [Kind]',
                                page_name,
                                'Missing Quantity [Kind] Definition Quantity'
                            )
                        )
            # Relation Connections
            if ivalue.get('QorQK') == self.mathmoddb['Quantity']:
                if any(
                    'MISSING RELATION TYPE' in val
                    for val in ivalue.get('RelationQQ', {}).values()
                ):
                    self.err.append(
                        self.error_message(
                            'Quantity [Kind]',
                            page_name,
                            'Missing Relation Type (Quantity [Kind])'
                        )
                    )
                if any(
                    'MISSING RELATION TYPE' in val
                    for val in ivalue.get('RelationQQK', {}).values()
                ):
                    self.err.append(
                        self.error_message(
                            'Quantity [Kind]',
                            page_name,
                            'Missing Relation Type (Quantity [Kind])'
                        )
                    )
            elif ivalue.get('QorQK') == self.mathmoddb['QuantityKind']:
                if any(
                    'MISSING RELATION TYPE' in val
                    for val in ivalue.get('RelationQKQK', {}).values()
                ):
                    self.err.append(
                        self.error_message(
                            'Quantity [Kind]',
                            page_name,
                            'Missing Relation Type (Quantity [Kind])'
                        )
                    )
                if any(
                    'MISSING RELATION TYPE' in val
                    for val in ivalue.get('RelationQKQ', {}).values()
                ):
                    self.err.append(
                        self.error_message(
                            'Quantity [Kind]',
                            page_name,
                            'Missing Relation Type (Quantity [Kind])'
                        )
                    )

    def problem(self, project, data):
        '''Perform Problem Checks:
            - Connections present
            - Relations present'''

        values = project.values.filter(
            snapshot = None,
            attribute = Attribute.objects.get(
                uri = f"{BASE_URI}domain/problem"
            )
        )

        if not data.get('problem'):
            self.err.append('Research Problem: No Research Problem Entered')
            return

        for ikey, ivalue in data.get('problem',{}).items():
            page_name = values.get(set_index=ikey).text
            # Check Connections
            if not ivalue.get('RelationRF'):
                self.err.append(
                    self.error_message(
                        'Research Problem',
                        page_name,
                        'Missing Academic Discipline'
                    )
                )
            # Relation Connections
            if any(
                'MISSING RELATION TYPE' in val
                for val in ivalue.get('RelationRP', {}).values()
            ):
                self.err.append(
                    self.error_message(
                        'Research Problem',
                        page_name,
                        'Missing Relation Type (Research Problem)'
                    )
                )

    def field(self, project, data):
        '''Perform Problem Checks:
            - Relations present'''

        values = project.values.filter(
            snapshot = None,
            attribute = Attribute.objects.get(
                uri = f"{BASE_URI}domain/field"
            )
        )

        if not data.get('field'):
            self.err.append('Academic Discipline: No Academic Discipline Entered')
            return

        for ikey, ivalue in data.get('field',{}).items():
            page_name = values.get(set_index=ikey).text
            # Relation Connections
            if any(
                'MISSING RELATION TYPE' in val
                for val in ivalue.get('RelationRF', {}).values()
            ):
                self.err.append(
                    self.error_message(
                        'Academic Discipline',
                        page_name,
                        'Missing Relation Type (Academic Discipline)'
                    )
                )

    def publication(self, project, data):
        '''Perform Problem Checks:
            - Relations present'''

        values = project.values.filter(
            snapshot = None,
            attribute = Attribute.objects.get(
                uri = f"{BASE_URI}domain/publication"
            )
        )

        if not data.get('publication'):
            self.err.append('Publication: No Publication Entered')
            return

        for ikey, ivalue in data.get('publication',{}).items():
            page_name = values.get(set_index=ikey).text
            if ivalue.get('ID') == 'not found' and not ivalue.get('reference'):
                self.err.append(
                    self.error_message(
                        'Publication',
                        page_name,
                        'Missing Publication DOI or URL'
                    )
                )
            # Check Connections
            if not ivalue.get('RelationP'):
                self.err.append(
                    self.error_message(
                        'Publication',
                        page_name,
                        'Missing Mathematical Model Entity'
                    )
                )
            # Relation Connections
            if any(
                'MISSING RELATION TYPE' in val
                for val in ivalue.get('RelationP', {}).values()
            ):
                self.err.append(
                    self.error_message(
                        'Publication',
                        page_name,
                        'Missing Relation Type (Publication)'
                    )
                )

    def run(self, project, data):
        '''Run All Checks'''
        self.id_name_description(project, data)
        self.properties(project, data)
        self.model(project, data)
        self.task(project, data)
        self.formulation(project, data)
        self.quantity(project, data)
        self.problem(project, data)
        self.field(project, data)
        self.publication(project, data)
        if self.err:
            self.err.sort()
            self.err.insert(
                0,
                "Following incomplete or inconsistent aspects prevented the export:"
            )
        return self.err



'''Module containing Utility Functions for the Algorithm Documentation'''

import re

from .constants import class_prefix_map

from ..config import BASE_URI
from ..helpers import value_editor
from ..getters import get_data, get_mathalgodb, get_options
from ..queries import query_sparql

def update_ids(project, ids, query, sparql_endpoint, source):
    """Update IDs of new MathAlgoDB Items and add them to the Questionnaire"""
    new_ids = {}

    for key, id_value in ids.items():
        # Ignore Items with MathAlgoDB ID
        if id_value.startswith(('mathmoddb:', 'bm:', 'pr:', 'so:', 'al:', 'pb')):
            continue

        # Get MathAlgoDB ID
        results = query_sparql(query.format(f'"{key}"'), sparql_endpoint)
        if not (results and results[0].get('ID', {}).get('value')):
            continue

        match = re.match(r"(\d+)(\D+)", id_value)
        if not match:
            continue

        set_index, set_name = match.groups()
        first_result = results[0]

        # Generate Entry
        value_editor(
            project=project,
            uri=f"{BASE_URI}domain/{set_name}/id",
            info = {
                'text': f"{key} ({first_result['quote']['value']}) [{source}]",
                'external_id': f"{source}:{first_result['ID']['value']}",
                'set_index': set_index
            }
        )

        if source == 'mathalgodb':
            class_value = first_result.get('class', {}).get('value')
            if not class_value:
                continue
            prefix = class_prefix_map.get(class_value)
            if prefix:
                new_ids[key] = f"{prefix}:{first_result['ID']['value']}"

    return new_ids

def dict_to_triples_mathalgodb(data):
    '''Turn Dictionary Entries into Subject-Predicate-Object Triples'''
    mathalgodb = get_mathalgodb()
    inverse_property_mapping = get_data('algorithm/data/inversePropertyMapping.json')
    options = get_options()

    relations = ['IntraClassRelation', 'P2A', 'P2B', 'P2S']
    relatants = ['IntraClassElement', 'ARelatant', 'BRelatant', 'SRelatant']

    triples = []
    ids = {}

    # Get ID Dict
    for idx, item in data.items():
        if item['ID'] and item['ID'].startswith('mathalgodb:'):
            _, mathalgodb_id = item['ID'].split(':')
            if 'algorithm' in idx:
                ids[item['Name']] = f"al:{mathalgodb_id}"
            if 'problem' in idx:
                ids[item['Name']] = f"pr:{mathalgodb_id}"
            if 'benchmark' in idx:
                ids[item['Name']] = f"bm:{mathalgodb_id}"
            if 'software' in idx:
                ids[item['Name']] = f"so:{mathalgodb_id}"
            if 'publication' in idx:
                ids[item['Name']] = f"pb:{mathalgodb_id}"
        else:
            ids[item['Name']] = idx

    # Go through all individuals
    for idx, item in data.items():

        # Get ID of Individual
        subject = ids[item['Name']]

        if not subject.startswith(("al:", "pr:", "so:", "pb:", "bm:", "mathalgodb:")):

            # Assign Individual Label
            triples.append((subject, "rdfs:label", f'"{item["Name"]}"'))

            # Assign Individual Description
            if item.get('Description'):
                if item['Description'] != 'No Description Provided!':
                    triples.append((subject, "rdfs:comment", f'"{item["Description"]}"'))

            # Assign Individual Class
            if 'algorithm' in idx:
                triples.append((subject, "a", 'mathalgodb:algorithm'))
            elif 'problem' in idx:
                triples.append((subject, "a", 'mathalgodb:problem'))
            elif 'software' in idx:
                triples.append((subject, "a", 'mathalgodb:software'))
            elif 'benchmark' in idx:
                triples.append((subject, "a", 'mathalgodb:benchmark'))
            elif 'publication' in idx:
                triples.append((subject, "a", 'mathalgodb:publication'))

            # Assign Individual References
            for reference in item.get('reference', {}).values():
                if reference[0] == options['DOI']:
                    doi_value = reference[1]
                    triples.append((subject, "dc:hasIdentifier", f'"doi:{doi_value}"'))
                if reference[0] == options['SWMATH']:
                    swmath_value = reference[1]
                    triples.append((subject, "dc:hasIdentifier", f'"swmath:{swmath_value}"'))
                if reference[0] == options['MORWIKI']:
                    morwiki_value = reference[1]
                    triples.append((subject, "dc:hasIdentifier", f'"morwiki:{morwiki_value}"'))
                if reference[0] == options['URL']:
                    url_value = reference[1]
                    triples.append((subject, "dc:hasIdentifier", f'"{url_value}"'))

        # Assign Individual Relations
        if 'algorithm' in idx:
            for relatant in ['PRelatant', 'SRelatant']:
                relatant_dict = item.get(relatant, {})
                for key in relatant_dict:
                    relation_uri = None
                    if relatant == 'PRelatant':
                        relation_uri = mathalgodb['solves']
                    elif relatant == 'SRelatant':
                        relation_uri = mathalgodb['implementedBy']
                    relatant_value = relatant_dict[key]
                    object_value = None
                    if relatant_value['ID'].startswith('mathalgodb:'):
                        _, mathalgodb_id = relatant_value['ID'].split(':')
                        if relatant == 'PRelatant':
                            object_value = f"pr:{mathalgodb_id}"
                        elif relation == 'SRelatant':
                            object_value = f"so:{mathalgodb_id}"
                    else:
                        referred_name = relatant_value['Name']
                        object_value = ids.get(referred_name)
                    triples.append(
                        (subject,
                         f"mathalgodb:{relation_uri.split('/')[-1]}",
                         object_value)
                    )
                    triples.append(
                        (object_value,
                         f"mathalgodb:{inverse_property_mapping[relation_uri].split('/')[-1]}",
                         subject)
                    )

        # Assign Individual Relations
        if 'problem' in idx:
            for relatant in ['BRelatant']:
                relatant_dict = item.get(relatant, {})
                for key in relatant_dict:
                    if relatant == 'BRelatant':
                        relation_uri = mathalgodb['instantiates']
                    relatant_value = relatant_dict[key]
                    if relatant_value['ID'].startswith('mathalgodb:'):
                        _, mathalgodb_id = relatant_value['ID'].split(':')
                        if relatant == 'BRelatant':
                            object_value = f"bm:{mathalgodb_id}"
                    else:
                        referred_name = relatant_value['Name']
                        object_value = ids.get(referred_name)
                    triples.append(
                        (subject,
                         f"mathalgodb:{relation_uri.split('/')[-1]}",
                         object_value)
                    )
                    triples.append(
                        (object_value,
                         f"mathalgodb:{inverse_property_mapping[relation_uri].split('/')[-1]}",
                         subject)
                    )

        # Assign Individual Relations
        if 'software' in idx:
            for relatant in ['BRelatant']:
                relatant_dict = item.get(relatant, {})
                for key in relatant_dict:
                    if relatant == 'BRelatant':
                        relation_uri = mathalgodb['tests']
                    relatant_value = relatant_dict[key]
                    if relatant_value['ID'].startswith('mathalgodb:'):
                        _, mathalgodb_id = relatant_value['ID'].split(':')
                        if relatant == 'BRelatant':
                            object_value = f"bm:{mathalgodb_id}"
                    else:
                        referred_name = relatant_value['Name']
                        object_value = ids.get(referred_name)
                    triples.append(
                        (subject,
                         f"mathalgodb:{relation_uri.split('/')[-1]}",
                         object_value)
                    )
                    triples.append(
                        (object_value,
                         f"mathalgodb:{inverse_property_mapping[relation_uri].split('/')[-1]}",
                         subject)
                    )

        # Assign Further Relations
        for relation, relatant in zip(relations,relatants):
            relation_dict = item.get(relation, {})
            relatant_dict = item.get(relatant, {})
            for key in relation_dict:
                if relatant_dict.get(key):
                    relation_uri = relation_dict[key]
                    relatant_value = relatant_dict[key]
                    if relatant_value['ID'].startswith('mathalgodb:'):
                        _, mathalgodb_id = relatant_value['ID'].split(':')
                        if relation == 'IntraClassRelation':
                            if 'algorithm' in idx:
                                object_value = f"al:{mathalgodb_id}"
                            if 'problem' in idx:
                                object_value = f"pr:{mathalgodb_id}"
                            if 'benchmark' in idx:
                                object_value = f"bm:{mathalgodb_id}"
                            if 'software' in idx:
                                object_value = f"so:{mathalgodb_id}"
                            if 'publication' in idx:
                                object_value = f"pb:{mathalgodb_id}"
                        elif relation == 'A2P':
                            object_value = f"pr:{mathalgodb_id}"
                        elif relation == 'A2S':
                            object_value = f"so:{mathalgodb_id}"
                        elif relation == 'P2B':
                            object_value = f"bm:{mathalgodb_id}"
                        elif relation == 'S2B':
                            object_value = f"bm:{mathalgodb_id}"
                        elif relation == 'P2A':
                            object_value = f"al:{mathalgodb_id}"
                        elif relation == 'P2B':
                            object_value = f"bm:{mathalgodb_id}"
                        elif relation == 'P2S':
                            object_value = f"so:{mathalgodb_id}"
                    else:
                        referred_name = relatant_value['Name']
                        object_value = ids.get(referred_name)
                    triples.append(
                        (subject,
                         f"mathalgodb:{relation_uri.split('/')[-1]}",
                         object_value)
                    )
                    triples.append(
                        (object_value,
                         f"mathalgodb:{inverse_property_mapping[relation_uri].split('/')[-1]}",
                         subject)
                    )

    return triples, ids

def generate_sparql_insert_with_new_ids_mathalgodb(triples):
    '''Generate SPARQL Insert Query'''
    # Step 1: Identify new items that need mardmo IDs
    new_items = {}
    counter = 0
    for triple in triples:
        subject = triple[0]
        if not subject.startswith(("al:", "pr:", "so:", "pb:", "bm:", "mathalgodb:")):
            # Assign temporary placeholders for new IDs
            new_items[subject] = f"newItem{counter}"
            counter += 1

    # Step 2: Generate SPARQL query with BIND for new mardmo IDs
    insert_query = """
    PREFIX mathalgodb: <https://mardi4nfdi.de/mathalgodb/0.1#>
    PREFIX bm: <https://mardi4nfdi.de/mathalgodb/0.1/benchmark#>
    PREFIX al: <https://mardi4nfdi.de/mathalgodb/0.1/algorithm#>
    PREFIX so: <https://mardi4nfdi.de/mathalgodb/0.1/software#>
    PREFIX pr: <https://mardi4nfdi.de/mathalgodb/0.1/problem#>
    PREFIX pb: <https://mardi4nfdi.de/mathalgodb/0.1/publication#>
    PREFIX rdfs: <http://www.w3.org/2000/01/rdf-schema#>
    PREFIX xsd: <http://www.w3.org/2001/XMLSchema#>
    PREFIX dc: <http://purl.org/spar/datacite/>

    INSERT{
    """
    # Construct the insert part
    for triple in triples:
        subject = triple[0]
        predicate = triple[1]
        obj = triple[2]

        # Replace new subjects with placeholders
        if subject in new_items:
            subject = f"?{new_items[subject]}"
        else:
            subject = f"{subject}"

        # Format object based on whether it's a literal or a URI
        if re.match(r'^https?://', obj):
            obj_formatted = f"<{obj}>"
        else:
            if obj.startswith(("al:", "pr:", "so:", "pb:", "bm:", "mathalgodb:", '"', ':', '<')):
                obj_formatted = f'{obj}'
            else:
                obj_formatted = f"?{new_items[obj]}"

        # Construct the triple in the query
        insert_query += f"  {subject} {predicate} {obj_formatted} .\n"

    insert_query += "}\nWHERE {\n"

    # Step 3: Add logic to get the next free mardmo ID
    insert_query += """
    {
      SELECT (MAX(?num) AS ?maxID) WHERE {
        ?id a ?type .
        FILTER (
          STRSTARTS(STR(?id), "https://mardi4nfdi.de/mathalgodb/0.1/benchmark#mardmo") ||
          STRSTARTS(STR(?id), "https://mardi4nfdi.de/mathalgodb/0.1/problem#mardmo") ||
          STRSTARTS(STR(?id), "https://mardi4nfdi.de/mathalgodb/0.1/software#mardmo") ||
          STRSTARTS(STR(?id), "https://mardi4nfdi.de/mathalgodb/0.1/algorithm#mardmo") ||
          STRSTARTS(STR(?id), "https://mardi4nfdi.de/mathalgodb/0.1/publication#mardmo")
        )
        BIND (xsd:integer(REPLACE(STR(?id), ".*#mardmo", "")) AS ?num)
      }
    }
    BIND (IF(BOUND(?maxID), ?maxID + 1, 0) AS ?nextID)
    """
    id_counter = 0
    for new_item in new_items:

        if 'algorithm' in new_item:
            insert_query += (
                "BIND(IRI(CONCAT('https://mardi4nfdi.de/mathalgodb/0.1/algorithm"
                f"#mardmo', STR(?nextID+{id_counter}))) AS ?{new_items[new_item]})\n"
            )
        elif 'benchmark' in new_item:
            insert_query += (
                "BIND(IRI(CONCAT('https://mardi4nfdi.de/mathalgodb/0.1/benchmark"
                f"#mardmo', STR(?nextID+{id_counter}))) AS ?{new_items[new_item]})\n"
            )
        elif 'software' in new_item:
            insert_query += (
                "BIND(IRI(CONCAT('https://mardi4nfdi.de/mathalgodb/0.1/software"
                f"#mardmo', STR(?nextID+{id_counter}))) AS ?{new_items[new_item]})\n"
            )
        elif 'problem' in new_item:
            insert_query += (
                "BIND(IRI(CONCAT('https://mardi4nfdi.de/mathalgodb/0.1/problem"
                f"#mardmo', STR(?nextID+{id_counter}))) AS ?{new_items[new_item]})\n"
            )
        elif 'publication' in new_item:
            insert_query += (
                "BIND(IRI(CONCAT('https://mardi4nfdi.de/mathalgodb/0.1/publication"
                f"#mardmo', STR(?nextID+{id_counter}))) AS ?{new_items[new_item]})\n"
            )

        id_counter += 1

    insert_query += "}"

    return insert_query



'''Module building handler maps for post_save and post_delete routing.'''

from .config import BASE_URI
from .getters import get_questions
from .model.handlers import Information as ModelInformation
from .algorithm.handlers import Information as AlgorithmInformation
from .workflow.handlers import Information as WorkflowInformation
from .publication.handlers import Information as PublicationInformation
from .handlers import Information as GeneralInformation

def build_handler_map():
    """Build a global mapping of attribute URIs to handler functions."""
    #from .model import handlers as model_handlers
    base = BASE_URI
    handler_map = {}

    # Questions
    questions_model = get_questions('model')
    questions_algorithm = get_questions('algorithm')
    questions_workflow = get_questions('workflow')
    questions_publication = get_questions('publication')

    # Information Classes
    model = ModelInformation()
    algorithm = AlgorithmInformation()
    workflow = WorkflowInformation()
    publication = PublicationInformation()
    general = GeneralInformation()

    # Model handlers
    handler_map.update({
        'mardmo-model-catalog': {
            f"{base}{questions_model['Research Field']['ID']['uri']}":
                model.field,
            f"{base}{questions_model['Research Problem']['ID']['uri']}":
                model.problem,
            f"{base}{questions_model['Quantity']['ID']['uri']}":
                model.quantity,
            f"{base}{questions_model['Mathematical Formulation']['ID']['uri']}":
                model.formulation,
            f"{base}{questions_model['Task']['ID']['uri']}":
                model.task,
            f"{base}{questions_model['Mathematical Model']['ID']['uri']}":
                model.model,
            f'{base}{questions_model["Task"]["QRelatant"]["uri"]}':
                general.relation,
            f'{base}{questions_model["Task"]["MFRelatant"]["uri"]}':
                general.relation,
            f'{base}{questions_model["Mathematical Formulation"]["Element Quantity"]["uri"]}':
                general.relation,
            f'{base}{questions_model["Quantity"]["Element Quantity"]["uri"]}':
                general.relation,
            f'{base}{questions_model["Mathematical Model"]["MFRelatant"]["uri"]}':
                general.relation,
            f'{base}{questions_model["Mathematical Model"]["Assumption"]["uri"]}':
                general.relation,
            f'{base}{questions_model["Task"]["Assumption"]["uri"]}':
                general.relation,
            f'{base}{questions_model["Mathematical Formulation"]["Assumption"]["uri"]}':
                general.relation,
            f'{base}{questions_model["Mathematical Formulation"]["MFRelatant"]["uri"]}':
                general.relation,
            f'{base}{questions_model["Research Problem"]["RFRelatant"]["uri"]}':
                general.relation,
            f'{base}{questions_model["Mathematical Model"]["RPRelatant"]["uri"]}':
                general.relation,
            f'{base}{questions_model["Mathematical Model"]["TRelatant"]["uri"]}':
                general.relation,
            f"{base}{questions_publication['Publication']['ID']['uri']}":
                publication.citation,
        }
    })

    # Algorithm handlers
    handler_map.update({
        'mardmo-algorithm-catalog': {
            f"{base}{questions_algorithm['Benchmark']['ID']['uri']}":
                algorithm.benchmark,
            f"{base}{questions_algorithm['Software']['ID']['uri']}":
                algorithm.software,
            f"{base}{questions_algorithm['Problem']['ID']['uri']}":
                algorithm.problem,
            f"{base}{questions_algorithm['Algorithm']['ID']['uri']}":
                algorithm.algorithm,
            f'{base}{questions_algorithm["Problem"]["BRelatant"]["uri"]}':
                general.relation,
            f'{base}{questions_algorithm["Software"]["BRelatant"]["uri"]}':
                general.relation,
            f'{base}{questions_algorithm["Algorithm"]["PRelatant"]["uri"]}':
                general.relation,
            f'{base}{questions_algorithm["Algorithm"]["SRelatant"]["uri"]}':
                general.relation,
            f"{base}{questions_publication['Publication']['ID']['uri']}":
                publication.citation,
        }
    })

    # Workflow handlers
    handler_map.update({
        'mardmo-interdisciplinary-workflow-catalog': {
            f"{base}{questions_workflow['Software']['ID']['uri']}":
                workflow.software,
            f"{base}{questions_workflow['Hardware']['ID']['uri']}":
                workflow.hardware,
            f"{base}{questions_workflow['Instrument']['ID']['uri']}":
                workflow.instrument,
            f"{base}{questions_workflow['Data Set']['ID']['uri']}":
                workflow.data_set,
            f"{base}{questions_workflow['Method']['ID']['uri']}":
                workflow.method,
            f"{base}{questions_workflow['Process Step']['ID']['uri']}":
                workflow.process_step,
            f'{base}{questions_workflow["Process Step"]["Input"]["uri"]}':
                general.relation,
            f'{base}{questions_workflow["Process Step"]["Output"]["uri"]}':
                general.relation,
            f'{base}{questions_workflow["Process Step"]["Method"]["uri"]}':
                general.relation,
            f'{base}{questions_workflow["Process Step"]["Environment-Software"]["uri"]}':
                general.relation,
            f'{base}{questions_workflow["Process Step"]["Environment-Instrument"]["uri"]}':
                general.relation,
            f'{base}{questions_workflow["Method"]["Software"]["uri"]}':
                general.relation,
            f'{base}{questions_workflow["Method"]["Instrument"]["uri"]}':
                general.relation,
            f'{base}{questions_workflow["Instrument"]["Software"]["uri"]}':
                general.relation,
            f'{base}{questions_workflow["Hardware"]["Software"]["uri"]}':
                general.relation,
            f"{base}{questions_publication['Publication']['ID']['uri']}":
                publication.citation,
        }
    })

    return handler_map



'''Functional Rules to feed User Answers in Dict'''

from .helpers import basic_dict, basic_list

def rule_0(value, attribute, config, _prefix_idx):
    "Handle Flag Combo 0 for Get Answers"
    entry = getattr(value, attribute)
    path = [config["key1"], config["key2"]]
    return entry, path

def rule_1(value, attribute, config, prefix_idx):
    "Handle Flag Combo 1 for Get Answers"
    entry = getattr(value, attribute)
    path = [config["key1"], prefix_idx, config["key2"]]
    return entry, path

def rule_2(value, attribute, config, _prefix_idx):
    "Handle Flag Combo 2 for Get Answers"
    entry = getattr(value, attribute)
    path = [config["key1"], value.set_index, config["key2"]]
    return entry, path

def rule_3(value, attribute, config, prefix_idx):
    "Handle Flag Combo 3 for Get Answers"
    entry = getattr(value, attribute)
    path = [config["key1"], prefix_idx, config["key2"], value.set_index]
    if config["key3"]:
        path.append(config["key3"])
    return entry, path

def rule_4(value, attribute, config, _prefix_idx):
    "Handle Flag Combo 4 for Get Answers"
    if attribute == 'text':
        entry = getattr(value, attribute)
    else:
        entry = basic_list(value)
    path = [config["key1"], value.set_index, config["key2"], value.collection_index]
    if config["key3"]:
        path.append(config["key3"])
    return entry, path

def rule_5(value, attribute, config, prefix_idx):
    "Handle Flag Combo 5 for Get Answers"
    if config["key2"] == 'reference':
        entry = basic_list(value)
    else:
        entry = getattr(value, attribute)
    path = [config["key1"], prefix_idx, config["key2"], value.collection_index]
    return entry, path

def rule_6(value, _attribute, config, prefix_idx):
    "Handle Flag Combo 6 for Get Answers"
    entry = basic_list(value)
    path = [config["key1"], prefix_idx, config["key2"], value.set_index, value.collection_index]
    return entry, path

def rule_7(value, _attribute, config, prefix_idx):
    "Handle Flag Combo 7 for Get Answers"
    entry = basic_list(value)
    path = [config["key1"], prefix_idx, config["key2"]]
    return entry, path

def rule_8(value, _attribute, config, _prefix_idx):
    "Handle Flag Combo 8 for Get Answers"
    entry = basic_list(value)
    path = [config["key1"], value.set_index, config["key2"]]
    return entry, path

def rule_9(value, _attribute, config, _prefix_idx):
    "Handle Flag Combo 9 for Get Answers"
    entry = basic_dict(value)
    path = [config["key1"]]
    return entry, path

def rule_10(value, attribute, config, _prefix_idx):
    "Handle Flag Combo 10 for Get Answers"
    entry = getattr(value, attribute)
    path = [config["key1"], config["key2"], value.collection_index]
    return entry, path

def rule_11(value, _attribute, config, _prefix_idx):
    "Handle Flag Combo 11 for Get Answers"
    entry = basic_dict(value)
    path = [config["key1"], config["key2"], value.collection_index]
    return entry, path

def rule_12(value, _attribute, config, _prefix_idx):
    "Handle Flag Combo 12 for Get Answers"
    entry = value.external_id
    path = [config["key1"], value.set_index, config["key2"]]
    return entry, path

def rule_13(value, _attribute, config, prefix_idx):
    "Handle Flag Combo 13 for Get Answers"
    if config["key2"] == 'DefinedQuantity':
        entry = basic_dict(value)
    else:
        entry = value.external_id
    path = [config["key1"], prefix_idx, config["key2"]]
    return entry, path

def rule_14(value, _attribute, config, prefix_idx):
    "Handle Flag Combo 14 for Get Answers"
    entry = basic_dict(value)
    path = [config["key1"], prefix_idx, config["key2"], value.set_index]
    if config["key3"]:
        path.append(config["key3"])
    return entry, path

def rule_15(value, _attribute, config, prefix_idx):
    "Handle Flag Combo 15 for Get Answers"
    entry = basic_dict(value)
    path = [config["key1"], prefix_idx,config["key2"], value.collection_index]
    return entry, path

def rule_16(value, _attribute, config, prefix_idx):
    "Handle Flag Combo 16 for Get Answers"
    entry = basic_dict(value)
    path = [config["key1"], prefix_idx, config["key2"], value.set_index, value.collection_index]
    return entry, path



'''Functions to create Payload for Export'''

from dataclasses import dataclass, field
from typing import Optional, Callable

import re

from .getters import get_items, get_properties
from .queries import query_item

@dataclass
class PayloadState:
    '''Data Class to store the current state of the Payload.'''
    counter: int = 0
    dictionary: dict = field(default_factory=dict)
    subject: dict = field(default_factory=dict)
    subject_item: Optional[str] = None

class GeneratePayload:
    '''Class to build the Payload for an Export to a Wikibase
       with Items, Statements, Qualifiers, and Checks.'''

    def __init__(self, url: str, items: dict,
                 relations: dict | None = None,
                 data_properties: Callable[[str], dict] | None = None):
        '''Instantiate Class Attributes'''
        # Input Attributes
        self.url: str = url
        self.user_items: dict = items
        self.relations: dict = relations
        self.data_properties: Callable[[str], dict] | None = data_properties
        self.properties: dict = get_properties()
        self.items: dict = get_items()
        # Working Attributes
        self.state: PayloadState = PayloadState()

    def _items_url(self):
        '''Get Item URL'''
        return f'{self.url}/w/rest.php/wikibase/v1/entities/items'

    def _statement_url(self, item):
        '''Get Statement URL'''
        return f'{self.url}/w/rest.php/wikibase/v1/entities/items/{item}/statements'

    def _build_item(self, identifier, label, description, statements = None):
        '''Build Item with ID, URL, Label, Description and optional Statement.'''
        # Empty Statements if none provided
        if statements is None:
            statements = []
        # Build Item
        item = {'id': identifier,
                'url': self._items_url(),
                'label': label,
                'description':  description,
                'statements': statements}
        return item

    def _build_statement(self, identifier, content, data_type = "wikibase-item", qualifiers = None):
        '''Build Statement with ID, Datatype, Content and optional Qualifiers.'''
        # Empty Qualifiers if none provided
        if qualifiers is None:
            qualifiers = []
        # Build Statement
        statement = {"statement":
                        {"property":
                            {"id": identifier,
                             "data_type": data_type},
                             "value":
                                {"type": "value",
                                 "content": content},
                             "qualifiers": qualifiers
                        }
                    }
        return statement

    def build_relation_check_query(self):
        '''Build SPARQL Check Query for Statement'''
        relation_keys = [k for k in self.state.dictionary if k.startswith('RELATION')]
        optional_blocks, bind_blocks = [], []

        for idx, key in enumerate(relation_keys):
            entry = self.state.dictionary[key]
            optional_block, bind_block = self._build_relation_block(idx, entry)
            if optional_block is None:
                continue
            optional_blocks.append(optional_block)
            bind_blocks.append(bind_block)

        query_body = '\n'.join(optional_blocks + bind_blocks)
        selectors = " ".join(f"?RELATION{idx}" for idx in range(len(relation_keys)))
        return f'\nSELECT {selectors} WHERE {{\n{query_body}\n}}'

    def _sparql_value(self, value, data_type):
        '''Format Value according to Data Type'''
        if data_type == 'wikibase-item':
            formatted_value = f'wd:{value}'
        elif data_type == 'string':
            formatted_value = f"'{value}'"
        elif data_type == 'quantity':
            formatted_value = f"'{value}'^^<http://www.w3.org/2001/XMLSchema#decimal>"
        elif data_type == 'time':
            formatted_value = f"'{value}'^^<http://www.w3.org/2001/XMLSchema#dateTime>"
        elif data_type == 'monolingualtext':
            formatted_value = f"'{value}'@en"
        elif data_type == 'math':
            escaped_value = value.replace("\\", "\\\\").replace("\"", "\\\"")
            formatted_value = f"'{escaped_value}'^^<http://www.w3.org/1998/Math/MathML>"
        else:
            formatted_value = f"'{value}'"
        return formatted_value

    def _build_qualifier_triples(self, qualifiers, idx):
        '''Build Qualifier Triples'''
        triples = ''
        for q in qualifiers:
            q_prop = q['property']['id']
            q_value = q['value']['content']
            q_data_type = q['property']['data_type']
            if q_value in self.state.dictionary and 'id' in self.state.dictionary[q_value]:
                q_value = self.state.dictionary[q_value]['id']
            triples += (
                f'    ?statement{idx} pq:{q_prop} '
                f'{self._sparql_value(q_value, q_data_type)} .\n'
            )
        return triples

    def _build_relation_block(self, idx, entry):
        '''Build Relation Blocks'''
        target_item_key = entry['url'].split('/')[-2]
        target_item_data = self.state.dictionary.get(target_item_key)
        if not target_item_data:
            return None, None

        target_item_id = target_item_data['id']
        statement = entry['payload']['statement']
        prop_id = statement['property']['id']
        value = statement['value']['content']
        data_type = statement['property']['data_type']

        if value in self.state.dictionary and 'id' in self.state.dictionary[value]:
            value = self.state.dictionary[value]['id']

        subject = f'wd:{target_item_id}'
        value_str = self._sparql_value(value, data_type)
        qualifiers = statement.get('qualifiers', [])
        qual_triples = self._build_qualifier_triples(qualifiers, idx)

        block = {
            'optional': (
                f'OPTIONAL {{\n'
                f'  {subject} p:{prop_id} ?statement{idx} .\n'
                f'  ?statement{idx} ps:{prop_id} {value_str} .\n'
                f'{qual_triples if qualifiers else ""}}}'
            ),
            'bind': f'BIND(BOUND(?statement{idx}) AS ?RELATION{idx})'
        }

        return block['optional'], block['bind']

    def _find_key_by_values(self, id_value, name_value, description_value):
        '''Find Key of Item by its Value'''
        for key, values in self.user_items.items():
            if (values['ID'] == id_value and
                values['Name'] == name_value and
                values['Description'] == description_value):
                return key
        return None

    def get_dictionary(self):
        '''Get a Dictionary'''
        # Get Target Dictionary
        target_dictionary = self.state.dictionary
        return target_dictionary

    def get_item_key(self, value, role='subject'):
        """Get the Key of an Item (Subject/Object)"""
        if not value.get('Name') or not value.get('Description'):
            raise ValueError("All Items need to have a 'Name' and 'Description'!")

        item_key = self._find_key_by_values(
            value['ID'],
            value['Name'],
            value['Description'],
        )

        if role == 'subject':
            self.state.subject = value
            self.state.subject_item = item_key

        return item_key

    def add_qualifier(self, identifier, data_type, content):
        '''Build Qualifier to used in Statement.'''
        # Build Qualifer
        qualifier = [{"property":
                        {"id": identifier,
                         "data_type": data_type},
                         "value": 
                            {"type": "value",
                             "content": content}
                    }]
        return qualifier

    def add_data_properties(self, item_class):
        '''Build Data Property Statements'''
        data_properties = self.data_properties(item_class)
        for prop in self.state.subject.get('Properties', {}).values():
            self.add_answer(
                verb=self.properties['instance of'],
                object_and_type=[
                    data_properties[prop],
                    'wikibase-item',
                ]
            )

    def add_check_results(self, check):
        '''Add Check Status to Statements'''
        relation_keys = [k for k in self.state.dictionary if k.startswith('RELATION')]
        for idx, key in enumerate(relation_keys):
            exists_key = f'RELATION{idx}'
            exists_value = check[0].get(exists_key, {}).get('value', 'false')
            self.state.dictionary[key]['exists'] = exists_value

    def add_answer(self, verb, object_and_type,
                   qualifier = None, subject = None):
        '''Add answer to Payload'''
        if subject is None:
            subject = self.state.subject_item
        if qualifier is None:
            qualifier = []
        # Gather Statement
        statement = {
            'property_id': verb,
            'datatype': object_and_type[1],
            'value': object_and_type[0]
        }
        # Pattern of New Item
        pattern = re.compile(r"^Item\d{10}$")
        # Add Relation
        if (
            self.state.dictionary[subject]['id']
            or (isinstance(statement['value'], str) and pattern.match(statement['value']))
        ):
            self._add_relation(
                item=subject,
                statement=statement,
                qualifier=qualifier
            )
        else:
            self._add_to_item_statement(
                item=subject,
                statement=statement,
                qualifier=qualifier
            )


    def add_answers(self, mardmo_property, wikibase_property, datatype = 'string'):
        '''Add answer to Payload.'''
        for entry in self.state.subject.get(mardmo_property, {}).values():
            if not re.search(r"<math.*?</math>", entry, re.DOTALL): #IGNORE MATHML EXPORT UNTIL HANDLERS RE-WRITTEN FOR LATEX
                self.add_answer(
                    verb=self.properties[wikibase_property],
                    object_and_type=[
                        entry,
                        datatype,
                    ]
                )

    def add_backward_relation(self, data, relation, relatants):
        '''Add backward relations to payload.'''
        for entry in data:
            for relatant in entry.get(relatants, {}).values():
                relatant_item = self.get_item_key(relatant, 'object')
                if self.state.subject_item == relatant_item:
                    entry_item = self.get_item_key(entry, 'object')
                    self.add_answer(
                        verb=relation,
                        object_and_type=[
                            entry_item,
                            'wikibase-item',
                        ]
                    )

    def add_forward_relation_single(self, relation, relatant,
                                    alternative = None, qualifier = None):
        '''Add single forward relation to payload.'''
        # Empty Qualifiers if none provided
        if qualifier is None:
            qualifier = []
        for entry in self.state.subject.get(relatant, {}).values():
            # Get Item Key
            entry_item = self.get_item_key(entry, 'object')
            if entry_item in self.state.dictionary:
                # Add Payload
                self.add_answer(
                    verb=relation,
                    object_and_type=[
                        entry_item,
                        'wikibase-item',
                    ],
                    qualifier=qualifier
                )
            else:
                # Add Payload
                self.add_answer(
                    verb=alternative['relation'],
                    object_and_type=[
                        entry.get(alternative['relatant']),
                        'string',
                    ],
                    qualifier=qualifier
                )

    def add_forward_relation_multiple(self, relation, relatant, reverse = False):
        '''Add multiple forward relations to payload.'''
        for key, prop in self.state.subject.get(relation, {}).items():
            # Get Item Key
            relatant_item = self.get_item_key(
                self.state.subject.get(relatant,{}).get(key,{}),
                'object'
            )
            # Get (optional) Order Number
            qualifier = []
            if self.state.subject.get('formulation_number'):
                qualifier = self.add_qualifier(
                    self.properties['series ordinal'],
                    'string', 
                    self.state.subject['formulation_number'][key]
                )
            # Get potential Qualifier
            if len(self.relations[prop]) == 2:
                if self.relations[prop][1] not in ('forward', 'backward'):
                    qualifier.extend(
                        self.add_qualifier(
                            self.properties['object has role'],
                            'wikibase-item',
                            self.relations[prop][1]
                        )
                    )
            # Add Payload
            if not reverse:
                self.add_answer(
                    verb=self.relations[prop][0],
                    object_and_type=[
                        relatant_item,
                        'wikibase-item',
                    ],
                    qualifier=qualifier
                )
            else:
                self.add_answer(
                    verb=self.relations[prop][0],
                    object_and_type=[
                        self.state.subject_item,
                        'wikibase-item',
                    ],
                    qualifier=qualifier,
                    subject=relatant_item
                )

    def add_in_defining_formula(self):
        '''Add in defining formula Statement'''
        for element in self.state.subject.get('element', {}).values():
            if not re.search(r"<math.*?</math>", element.get('symbol', ''), re.DOTALL): #IGNORE MATHML EXPORT UNTIL HANDLERS RE-WRITTEN FOR LATEX
                # Get Item Key
                quantity_item = self.get_item_key(
                    element.get('quantity', {}),
                    'object'
                )
                # Add Quantity Qualifier
                qualifier = self.add_qualifier(
                    self.properties['symbol represents'],
                    'wikibase-item',
                    quantity_item
                )
                # Add Symbol to Payload
                if self.state.subject_item == quantity_item:
                    self._add_relation(
                        item = self.state.subject_item,
                        statement = {
                            'property_id': self.properties['in defining formula'],
                            'value': element.get('symbol', ''),
                            'datatype': 'math'
                        },
                        qualifier = qualifier
                    )
                else:
                    self.add_answer(
                        verb=self.properties['in defining formula'],
                        object_and_type=[
                            element.get('symbol', ''),
                            'math',
                        ],
                        qualifier=qualifier
                    )

    def _add_entry(self, key, value):
        '''Add Entry to Payload'''
        self.state.dictionary[key] = value

    def _add_to_item_statement(self, item, statement, qualifier=None):
        '''Add to Statement of Item'''
        if qualifier is None:
            qualifier = []
        self.state.dictionary[item]['statements'].append(
            [
                statement['property_id'],
                statement['datatype'],
                statement['value'],
                qualifier
            ]
        )

    def _add_relation(self, item, statement, qualifier=None):
        '''Add Statement to Item'''
        if qualifier is None:
            qualifier = []
        key = f"RELATION{self.state.counter}"
        self.state.dictionary[key] = {
            'id': '',
            'url': self._statement_url(item),
            'payload': self._build_statement(
                statement['property_id'],
                statement['value'],
                statement['datatype'],
                qualifier
            )
        }
        self.state.counter += 1

    def add_intra_class_relation(self, relation, relatant):
        '''Add forward/backward relations to Items of same Class'''
        for key, prop in self.state.subject.get(relation, {}).items():
            # Get Item Key
            relatant_item = self.get_item_key(
                self.state.subject.get(relatant, {}).get(key),
                'object'
            )
            # Add potential Qualifier
            qualifier = []
            if self.state.subject.get('assumption', {}).get(key):
                for assumption in self.state.subject['assumption'][key].values():
                    assumption_item = self.get_item_key(
                        assumption,
                        'object'
                    )
                    qualifier.extend(
                        self.add_qualifier(
                            self.properties['assumes'],
                            'wikibase-item',
                            assumption_item
                        )
                    )
            if self.state.subject.get('task_number', {}).get(key):
                qualifier.extend(
                    self.add_qualifier(
                        self.properties['series ordinal'],
                        'string',
                        self.state.subject['task_number'][key]
                    )
                )
            # Add Forward or Backward Relation
            if self.relations[prop][1] == 'forward':
                self.add_answer(
                    verb=self.relations[prop][0],
                    object_and_type=[
                        relatant_item,
                        'wikibase-item',
                    ],
                    qualifier=qualifier
                )
            elif self.relations[prop][1] == 'backward':
                self.add_answer(
                    verb=self.relations[prop][0],
                    object_and_type=[
                        self.state.subject_item,
                        'wikibase-item',
                    ],
                    qualifier=qualifier,
                    subject=relatant_item
                )

    def add_item_payload(self):
        '''Add Payload String to Item'''
        for item_id, item_data in self.state.dictionary.items():
            # Check if Item in Payload
            if not item_id.startswith('Item'):
                continue
            # Extract Information
            label = item_data.get("label", "")
            description = item_data.get("description", "")
            statements_input = item_data.get("statements", [])
            # Grouped statements by PID
            statements = {}
            for s in statements_input:
                pid, dtype, obj = s[0], s[1], s[2]
                qualifier = None
                if len(s) == 4:
                    qualifier = s[3]
                statement = {
                    "property": {"id": pid, "data_type": dtype},
                    "value": {"type": "value", "content": obj}
                }
                if qualifier:
                    statement["qualifiers"] = qualifier

                statements.setdefault(pid, []).append(statement)
            # Build payload
            payload = {
                "item": {
                    "labels": {"en": label},
                    "statements": statements
                }
            }
            if description:
                payload["item"]["descriptions"] = {"en": description}
            # Attach to original dict
            item_data["payload"] = payload

    def _check_mardi_and_raise(self, name: str, description: str):
        """Check if item exists in MaRDI Portal and raise error if it does."""
        mardi_identifier = query_item(name, description)
        if mardi_identifier:
            raise ValueError(
                f"An item ({mardi_identifier}) with the label '{name}' "
                f"and description '{description}' already exists on the MaRDI Portal. "
                "If you intend to use this item, please select it in the questionnaire. "
                "Otherwise, redefine it."
            )
        return mardi_identifier

    def _statement_by_id_type(self, value: dict, id_type: str):
        """Return statements for a user-defined item based on ID type."""
        statements = []
        if id_type == 'wikidata':
            # Add Wikidata ID
            statements.append(
                [
                    self.properties['Wikidata QID'],
                    'external-id',
                    value['ID'].split(':')[1]
                ]
            )
        if id_type == 'no author found':
            # Add ORCID ID Statement
            if value.get('orcid'):
                statements.append(
                    [
                        self.properties['ORCID iD'],
                        'external-id',
                        value['orcid']
                    ]
                )
            # Add zbMath ID Statement
            if value.get('zbmath'):
                statements.append(
                    [
                        self.properties['zbMATH author ID'],
                        'external-id',
                        value['zbmath']
                    ]
                )
            # If Authors has ID, add further Statements
            if statements:
                statements.append(
                    [
                        self.properties['instance of'],
                        'wikibase-item',
                        self.items['human']
                    ]
                )
        if id_type == 'no journal found':
            # Add ISSN ID Statement
            if value.get('issn'):
                statements.append(
                    [
                        self.properties['ISSN'],
                        'external-id',
                        value['issn']
                    ]
                )
            # Add further Statements
            statements.append(
                [
                    self.properties['instance of'],
                    'wikibase-item',
                    self.items['scientific journal']
                ]
            )
        return statements

    def process_items(self):
        """Process Items"""
        handlers = {
            'mardi': lambda key, value: 
                self._add_entry(
                    key,
                    self._build_item(value['ID'].split(':')[1],
                    value['Name'],
                    value['Description']
                )
            ),
            'wikidata': lambda key, value: (
                self._check_mardi_and_raise(
                    value['Name'],
                    value['Description']
                ),
                self._add_entry(
                    key,
                    self._build_item(
                        '',
                        value['Name'],
                        value['Description'],
                        self._statement_by_id_type(
                            value,
                            'wikidata'))
                )
            ),
            'mathalgodb': lambda key, value: (
                self._check_mardi_and_raise(
                    value['Name'],
                    value['Description']
                ),
                self._add_entry(
                    key,
                    self._build_item(
                        '',
                        value['Name'],
                        value['Description']
                    )
                )
            ),
            'not found': lambda key, value: (
                self._check_mardi_and_raise(
                    value['Name'],
                    value['Description']
                ),
                self._add_entry(
                    key,
                    self._build_item(
                        '',
                        value['Name'],
                        value['Description'],
                        self._statement_by_id_type(
                            value,
                            'not found'
                        )
                    )
                )
            ),
            'no author found': lambda key, value: (
                self._check_mardi_and_raise(
                    value['Name'],
                    value['Description']
                ),
                self._add_entry(
                    key,
                    self._build_item(
                        '',
                        value['Name'],
                        value['Description'],
                        self._statement_by_id_type(
                            value,
                            'no author found'
                        )
                    )
                )
            ),
            'no journal found': lambda key, value: (
                self._check_mardi_and_raise(
                    value['Name'],
                    value['Description']
                ),
                self._add_entry(
                    key,
                    self._build_item(
                        '',
                        value['Name'],
                        value['Description'],
                        self._statement_by_id_type(
                            value,
                            'no journal found'
                        )
                    )
                )
            ),
        }

        for key, value in self.user_items.items():
            if not value.get('ID'):
                continue
            for id_type, handler in handlers.items():
                if id_type in value['ID']:
                    if id_type == 'no author found':
                        if value['zbmath'] or value['orcid']:
                            handler(key, value)
                    else:
                        handler(key, value)
                    break
